"""Get the LLM client."""
from enum import Enum
from cognee.infrastructure.llm import get_llm_config

# Define an Enum for LLM Providers
class LLMProvider(Enum):
    OPENAI = "openai"
    OLLAMA = "ollama"
    ANTHROPIC = "anthropic"
    CUSTOM = "custom"

def get_llm_client():
    """Get the LLM client based on the configuration using Enums."""
    llm_config = get_llm_config()

    provider = LLMProvider(llm_config.llm_provider)

    if provider == LLMProvider.OPENAI:
        if llm_config.llm_api_key is None:
            raise ValueError("LLM API key is not set.")

        from .openai.adapter import OpenAIAdapter
        return OpenAIAdapter(api_key=llm_config.llm_api_key, model=llm_config.llm_model, transcription_model=llm_config.transcription_model, streaming=llm_config.llm_streaming)
    elif provider == LLMProvider.OLLAMA:
        if llm_config.llm_api_key is None:
            raise ValueError("LLM API key is not set.")

        from .generic_llm_api.adapter import GenericAPIAdapter
        return GenericAPIAdapter(llm_config.llm_endpoint, llm_config.llm_api_key, llm_config.llm_model, "Ollama")
    elif provider == LLMProvider.ANTHROPIC:
        from .anthropic.adapter import AnthropicAdapter
        return AnthropicAdapter(llm_config.llm_model)
    elif provider == LLMProvider.CUSTOM:
        if llm_config.llm_api_key is None:
            raise ValueError("LLM API key is not set.")

        from .generic_llm_api.adapter import GenericAPIAdapter
        return GenericAPIAdapter(llm_config.llm_endpoint, llm_config.llm_api_key, llm_config.llm_model, "Custom")
    else:
        raise ValueError(f"Unsupported LLM provider: {provider}")



from typing import BinaryIO
import filetype
from .is_text_content import is_text_content

class FileTypeException(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message

class TxtFileType(filetype.Type):
    """Text file type"""
    MIME = "text/plain"
    EXTENSION = "txt"

    def __init__(self):
        super(TxtFileType, self).__init__(mime = TxtFileType.MIME, extension = TxtFileType.EXTENSION)

    def match(self, buf):
        return is_text_content(buf)

txt_file_type = TxtFileType()

filetype.add_type(txt_file_type)

class CustomPdfMatcher(filetype.Type):
    MIME = "application/pdf"
    EXTENSION = "pdf"

    def __init__(self):
        super(CustomPdfMatcher, self).__init__(mime = CustomPdfMatcher.MIME, extension = CustomPdfMatcher.EXTENSION)

    def match(self, buf):
        return b"PDF-" in buf

custom_pdf_matcher = CustomPdfMatcher()

filetype.add_type(custom_pdf_matcher)


def guess_file_type(file: BinaryIO) -> filetype.Type:
    file_type = filetype.guess(file)

    if file_type is None:
        raise FileTypeException("Unknown file detected.")

    return file_type



from datetime import datetime, timezone
from sqlalchemy import Column, DateTime, UUID, ForeignKey
from cognee.infrastructure.databases.relational import Base

class PipelineTask(Base):
    __tablename__ = "pipeline_task"

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))

    pipeline_id = Column("pipeline", UUID(as_uuid = True), ForeignKey("pipeline.id"), primary_key = True)
    task_id = Column("task", UUID(as_uuid = True), ForeignKey("task.id"), primary_key = True)



from typing import List, Union, Literal, Optional
from pydantic import BaseModel

class BaseClass(BaseModel):
    id: str
    name: str
    type: Literal["Class"] = "Class"
    description: str
    constructor_parameters: Optional[List[str]]

class Class(BaseModel):
    id: str
    name: str
    type: Literal["Class"] = "Class"
    description: str
    constructor_parameters: Optional[List[str]]
    from_class: Optional[BaseClass]

class ClassInstance(BaseModel):
    id: str
    name: str
    type: Literal["ClassInstance"] = "ClassInstance"
    description: str
    from_class: Class

class Function(BaseModel):
    id: str
    name: str
    type: Literal["Function"] = "Function"
    description: str
    parameters: Optional[List[str]]
    return_type: str
    is_static: Optional[bool] = False

class Variable(BaseModel):
    id: str
    name: str
    type: Literal["Variable"] = "Variable"
    description: str
    is_static: Optional[bool] = False
    default_value: Optional[str]

class Operator(BaseModel):
    id: str
    name: str
    type: Literal["Operator"] = "Operator"
    description: str
    return_type: str

class ExpressionPart(BaseModel):
    id: str
    name: str
    type: Literal["Expression"] = "Expression"
    description: str
    expression: str
    members: List[Union[Variable, Function, Operator]]

class Expression(BaseModel):
    id: str
    name: str
    type: Literal["Expression"] = "Expression"
    description: str
    expression: str
    members: List[Union[Variable, Function, Operator, ExpressionPart]]

class Edge(BaseModel):
    source_node_id: str
    target_node_id: str
    relationship_name: Literal["called in", "stored in", "defined in", "returned by", "instantiated in", "uses", "updates"]

class SourceCodeGraph(BaseModel):
    id: str
    name: str
    description: str
    language: str
    nodes: List[Union[
        Class,
        Function,
        Variable,
        Operator,
        Expression,
        ClassInstance,
    ]]
    edges: List[Edge]



from typing import BinaryIO
from pypdf import PdfReader

def extract_text_from_file(file: BinaryIO, file_type) -> str:
    """Extract text from a file"""
    if file_type.extension == "pdf":
        reader = PdfReader(stream = file)
        pages = list(reader.pages[:3])
        return "\n".join([page.extract_text().strip() for page in pages])

    if file_type.extension == "txt":
        return file.read().decode("utf-8")



import inspect
import logging
from ..tasks.Task import Task

logger = logging.getLogger("run_tasks(tasks: [Task], data)")

async def run_tasks(tasks: [Task], data = None):
    if len(tasks) == 0:
        yield data
        return

    args = [data] if data is not None else []

    running_task = tasks[0]
    leftover_tasks = tasks[1:]
    next_task = leftover_tasks[0] if len(leftover_tasks) > 1 else None
    next_task_batch_size = next_task.task_config["batch_size"] if next_task else 1

    if inspect.isasyncgenfunction(running_task.executable):
        logger.info("Running async generator task: `%s`", running_task.executable.__name__)
        try:
            results = []

            async_iterator = running_task.run(*args)

            async for partial_result in async_iterator:
                results.append(partial_result)

                if len(results) == next_task_batch_size:
                    async for result in run_tasks(leftover_tasks, results[0] if next_task_batch_size == 1 else results):
                        yield result

                    results = []

            if len(results) > 0:
                async for result in run_tasks(leftover_tasks, results):
                    yield result

                results = []

            logger.info("Finished async generator task: `%s`", running_task.executable.__name__)
        except Exception as error:
            logger.error(
                "Error occurred while running async generator task: `%s`\n%s\n",
                running_task.executable.__name__,
                str(error),
                exc_info = True,
            )
            raise error

    elif inspect.isgeneratorfunction(running_task.executable):
        logger.info("Running generator task: `%s`", running_task.executable.__name__)
        try:
            results = []

            for partial_result in running_task.run(*args):
                results.append(partial_result)

                if len(results) == next_task_batch_size:
                    async for result in run_tasks(leftover_tasks, results[0] if next_task_batch_size == 1 else results):
                        yield result

                    results = []

            if len(results) > 0:
                async for result in run_tasks(leftover_tasks, results):
                    yield result

                results = []

            logger.info("Finished generator task: `%s`", running_task.executable.__name__)
        except Exception as error:
            logger.error(
                "Error occurred while running generator task: `%s`\n%s\n",
                running_task.executable.__name__,
                str(error),
                exc_info = True,
            )
            raise error

    elif inspect.iscoroutinefunction(running_task.executable):
        logger.info("Running coroutine task: `%s`", running_task.executable.__name__)
        try:
            task_result = await running_task.run(*args)

            async for result in run_tasks(leftover_tasks, task_result):
                yield result

            logger.info("Finished coroutine task: `%s`", running_task.executable.__name__)
        except Exception as error:
            logger.error(
                "Error occurred while running coroutine task: `%s`\n%s\n",
                running_task.executable.__name__,
                str(error),
                exc_info = True,
            )
            raise error
                
    elif inspect.isfunction(running_task.executable):
        logger.info("Running function task: `%s`", running_task.executable.__name__)
        try:
            task_result = running_task.run(*args)

            async for result in run_tasks(leftover_tasks, task_result):
                yield result

            logger.info("Finished function task: `%s`", running_task.executable.__name__)
        except Exception as error:
            logger.error(
                "Error occurred while running function task: `%s`\n%s\n",
                running_task.executable.__name__,
                str(error),
                exc_info = True,
            )
            raise error



from .config import get_relational_config
from .create_relational_engine import create_relational_engine


def get_relational_engine():
    relational_config = get_relational_config()

    return create_relational_engine(**relational_config.to_dict())





from pydantic import BaseModel
from typing import List, Optional, Dict, Any, Union

class Relationship(BaseModel):
    type: str
    attributes: Optional[Dict[str, Any]] = {}

class Document(BaseModel):
    name: str
    content: str
    filetype: str

class Directory(BaseModel):
    name: str
    documents: List[Document] = []
    directories: List['Directory'] = []

# Allows recursive Directory Model
Directory.update_forward_refs()

class RepositoryProperties(BaseModel):
    custom_properties: Optional[Dict[str, Any]] = None
    location: Optional[str] = None  # Simplified location reference

class RepositoryNode(BaseModel):
    node_id: str
    node_type: str  # 'document' or 'directory'
    properties: RepositoryProperties = RepositoryProperties()
    content: Union[Document, Directory, None] = None
    relationships: List[Relationship] = []

class RepositoryGraphModel(BaseModel):
    root: RepositoryNode
    default_relationships: List[Relationship] = []



from typing import Any, Callable, Generator
import asyncio
from ..tasks.Task import Task

def run_tasks_parallel(tasks: [Task]) -> Callable[[Any], Generator[Any, Any, Any]]:
    async def parallel_run(*args, **kwargs):
        parallel_tasks = [asyncio.create_task(task.run(*args, **kwargs)) for task in tasks]

        results = await asyncio.gather(*parallel_tasks)
        return results[len(results) - 1] if len(results) > 1 else []

    return Task(parallel_run)



from enum import Enum
from typing import List

from pydantic import BaseModel



class TextSubclass(str, Enum):
    SOURCE_CODE = "Source code in various programming languages"
    SHELL_SCRIPTS = "Shell commands and scripts"
    MARKUP_LANGUAGES = "Markup languages (HTML, XML)"
    STYLESHEETS = "Stylesheets (CSS) and configuration files (YAML, JSON, INI)"
    OTHER = "Other that does not fit into any of the above categories"

class ContentType(BaseModel):
    """Base class for content type, storing type of content as string."""
    type: str = "TEXT"

class TextContent(ContentType):
    """Textual content class for more specific text categories."""
    subclass: List[TextSubclass]

class CodeContentPrediction(BaseModel):
    """Model to predict the type of content."""
    label: TextContent


from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy import Column, UUID, DateTime, String, JSON
from cognee.infrastructure.databases.relational import Base

class TaskRun(Base):
    __tablename__ = "task_runs"

    id = Column(UUID(as_uuid = True), primary_key = True, default = uuid4)

    task_name = Column(String)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))

    status = Column(String)

    run_info = Column(JSON)



from typing import List, Optional
from fastembed import TextEmbedding
from cognee.root_dir import get_absolute_path
from cognee.infrastructure.databases.vector.embeddings.EmbeddingEngine import EmbeddingEngine

class FastembedEmbeddingEngine(EmbeddingEngine):
    embedding_model: str
    embedding_dimensions: int

    def __init__(
        self,
        embedding_model: Optional[str] = "BAAI/bge-large-en-v1.5",
        embedding_dimensions: Optional[int] = 1024,
    ):
        self.embedding_model = embedding_model
        self.embedding_dimensions = embedding_dimensions

    async def embed_text(self, text: List[str]) -> List[float]:
        embedding_model = TextEmbedding(model_name = self.embedding_model, cache_dir = get_absolute_path("cache/embeddings"))
        embeddings_list = list(map(lambda embedding: embedding.tolist(), embedding_model.embed(text)))

        return embeddings_list

    def get_vector_size(self) -> int:
        return self.embedding_dimensions



from sqlalchemy import select
from cognee.infrastructure.databases.relational import get_relational_engine
from ..models import Dataset

async def retrieve_datasets(dataset_names: list[str]) -> list[Dataset]:
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        datasets = (await session.scalars(
            select(Dataset).filter(Dataset.name.in_(dataset_names))
        )).all()

        return datasets



from typing import List, Union, BinaryIO
from os import path
import asyncio
import dlt
import duckdb

import cognee.modules.ingestion as ingestion
from cognee.infrastructure.files.storage import LocalStorage
from cognee.modules.ingestion import get_matched_datasets, save_data_to_file
from cognee.shared.utils import send_telemetry
from cognee.base_config import get_base_config
from cognee.infrastructure.databases.relational import get_relational_config, get_relational_engine, create_db_and_tables
from cognee.modules.users.methods import create_default_user, get_default_user
from cognee.modules.users.permissions.methods import give_permission_on_document
from cognee.modules.users.models import User
from cognee.modules.data.operations.ensure_dataset_exists import ensure_dataset_exists

async def add(data: Union[BinaryIO, List[BinaryIO], str, List[str]], dataset_name: str = "main_dataset", user: User = None):
    await create_db_and_tables()

    if isinstance(data, str):
        if "data://" in data:
            # data is a data directory path
            datasets = get_matched_datasets(data.replace("data://", ""), dataset_name)
            return await asyncio.gather(*[add(file_paths, dataset_name) for [dataset_name, file_paths] in datasets])

        if "file://" in data:
            # data is a file path
            return await add([data], dataset_name)

        # data is text
        else:
            file_path = save_data_to_file(data, dataset_name)
            return await add([file_path], dataset_name)

    if hasattr(data, "file"):
        file_path = save_data_to_file(data.file, dataset_name, filename = data.filename)
        return await add([file_path], dataset_name)

    # data is a list of file paths or texts
    file_paths = []

    for data_item in data:
        if hasattr(data_item, "file"):
            file_paths.append(save_data_to_file(data_item, dataset_name, filename = data_item.filename))
        elif isinstance(data_item, str) and (
            data_item.startswith("/") or data_item.startswith("file://")
        ):
            file_paths.append(data_item)
        elif isinstance(data_item, str):
            file_paths.append(save_data_to_file(data_item, dataset_name))

    if len(file_paths) > 0:
        return await add_files(file_paths, dataset_name, user)

    return []

async def add_files(file_paths: List[str], dataset_name: str,  user):
    base_config = get_base_config()
    data_directory_path = base_config.data_root_directory

    processed_file_paths = []

    for file_path in file_paths:
        file_path = file_path.replace("file://", "")

        if data_directory_path not in file_path:
            file_name = file_path.split("/")[-1]
            file_directory_path = data_directory_path + "/" + (dataset_name.replace(".", "/") + "/" if dataset_name != "main_dataset" else "")
            dataset_file_path = path.join(file_directory_path, file_name)

            LocalStorage.ensure_directory_exists(file_directory_path)

            LocalStorage.copy_file(file_path, dataset_file_path)
            processed_file_paths.append(dataset_file_path)
        else:
            processed_file_paths.append(file_path)

    relational_config = get_relational_config()

    if relational_config.db_provider == "duckdb":
        db = duckdb.connect(relational_config.db_file_path)

        destination = dlt.destinations.duckdb(
          credentials = db,
        )
    else:
        destination = dlt.destinations.postgres(
            credentials = {
                "host": relational_config.db_host,
                "port": relational_config.db_port,
                "user": relational_config.db_user,
                "password": relational_config.db_password,
                "database": relational_config.db_name,
            },
        )

    pipeline = dlt.pipeline(
        pipeline_name = "file_load_from_filesystem",
        destination = destination,
    )

    dataset_name = dataset_name.replace(" ", "_").replace(".", "_") if dataset_name is not None else "main_dataset"
    dataset = await ensure_dataset_exists(dataset_name)

    @dlt.resource(standalone = True, merge_key = "id")
    async def data_resources(file_paths: str, user: User):
        for file_path in file_paths:
            with open(file_path.replace("file://", ""), mode = "rb") as file:
                classified_data = ingestion.classify(file)

                data_id = ingestion.identify(classified_data)

                file_metadata = classified_data.get_metadata()

                from sqlalchemy import select
                from cognee.modules.data.models import Data
                db_engine = get_relational_engine()
                async with db_engine.get_async_session() as session:
                    data = (await session.execute(
                        select(Data).filter(Data.id == data_id)
                    )).scalar_one_or_none()

                    if data is not None:
                        data.name = file_metadata["name"]
                        data.raw_data_location = file_metadata["file_path"]
                        data.extension = file_metadata["extension"]
                        data.mime_type = file_metadata["mime_type"]

                        await session.merge(data)
                        await session.commit()
                    else:
                        data = Data(
                            id = data_id,
                            name = file_metadata["name"],
                            raw_data_location = file_metadata["file_path"],
                            extension = file_metadata["extension"],
                            mime_type = file_metadata["mime_type"],
                        )
                        dataset.data.append(data)

                        await session.merge(dataset)

                        await session.commit()

                yield {
                    "id": data_id,
                    "name": file_metadata["name"],
                    "file_path": file_metadata["file_path"],
                    "extension": file_metadata["extension"],
                    "mime_type": file_metadata["mime_type"],
                }

                await give_permission_on_document(user, data_id, "read")
                await give_permission_on_document(user, data_id, "write")


    if user is None:
        user = await get_default_user()

        if user is None:
            user = await create_default_user()

    run_info = pipeline.run(
        data_resources(processed_file_paths, user),
        table_name = "file_metadata",
        dataset_name = dataset_name,
        write_disposition = "merge",
    )
    send_telemetry("cognee.add")

    return run_info



from cognee.infrastructure.llm import get_llm_config
from .EmbeddingEngine import EmbeddingEngine
from .LiteLLMEmbeddingEngine import LiteLLMEmbeddingEngine

def get_embedding_engine() -> EmbeddingEngine:
    llm_config = get_llm_config()
    return LiteLLMEmbeddingEngine(api_key = llm_config.llm_api_key)



from fastapi.security import OAuth2PasswordRequestForm
from fastapi_users.exceptions import UserNotExists
from cognee.infrastructure.databases.relational import get_relational_engine
from ...get_user_manager import get_user_manager_context
from ...get_user_db import get_user_db_context

async def authenticate_user(email: str, password: str):
    try:
        relational_engine = get_relational_engine()

        async with relational_engine.get_async_session() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    credentials = OAuth2PasswordRequestForm(username = email, password = password)
                    user = await user_manager.authenticate(credentials)
                    if user is None or not user.is_active:
                        return None
                    return user
    except UserNotExists as error:
        print(f"User {email} doesn't exist")
        raise error



# Don't add any more code here, this file is used only for the purpose
# of enabling imports from `cognee.pipelines` module.
# `from cognee.pipelines import Task` for example.

from .modules.pipelines import *



import os
from cognee.modules.data.processing.document_types.PdfDocument import PdfDocument

if __name__ == "__main__":
    test_file_path = os.path.join(os.path.dirname(__file__), "artificial-inteligence.pdf")
    pdf_doc = PdfDocument("Test document.pdf", test_file_path, chunking_strategy="paragraph")
    reader = pdf_doc.get_reader()

    for paragraph_data in reader.read():
        print(paragraph_data["word_count"])
        print(paragraph_data["text"])
        print(paragraph_data["cut_type"])
        print("\n")



from uuid import UUID, uuid5, NAMESPACE_OID

from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk
from cognee.tasks.chunking import chunk_by_paragraph

class TextChunker():
    id: UUID
    max_chunk_size: int

    chunk_index = 0
    chunk_size = 0
    paragraph_chunks = []

    def __init__(self, id: UUID, get_text: callable, chunk_size: int = 1024):
        self.id = id
        self.max_chunk_size = chunk_size
        self.get_text = get_text

    def read(self):
        for content_text in self.get_text():
            for chunk_data in chunk_by_paragraph(
                content_text,
                self.max_chunk_size,
                batch_paragraphs = True,
            ):
                if self.chunk_size + chunk_data["word_count"] <= self.max_chunk_size:
                    self.paragraph_chunks.append(chunk_data)
                    self.chunk_size += chunk_data["word_count"]
                else:
                    if len(self.paragraph_chunks) == 0:
                        yield DocumentChunk(
                            text = chunk_data["text"],
                            word_count = chunk_data["word_count"],
                            document_id = str(self.id),
                            chunk_id = str(chunk_data["chunk_id"]),
                            chunk_index = self.chunk_index,
                            cut_type = chunk_data["cut_type"],
                        )
                        self.paragraph_chunks = []
                        self.chunk_size = 0
                    else:
                        chunk_text = " ".join(chunk["text"] for chunk in self.paragraph_chunks)
                        yield DocumentChunk(
                            text = chunk_text,
                            word_count = self.chunk_size,
                            document_id = str(self.id),
                            chunk_id = str(uuid5(NAMESPACE_OID, f"{str(self.id)}-{self.chunk_index}")),
                            chunk_index = self.chunk_index,
                            cut_type = self.paragraph_chunks[len(self.paragraph_chunks) - 1]["cut_type"],
                        )
                        self.paragraph_chunks = [chunk_data]
                        self.chunk_size = chunk_data["word_count"]

                    self.chunk_index += 1

        if len(self.paragraph_chunks) > 0:
            yield DocumentChunk(
                text = " ".join(chunk["text"] for chunk in self.paragraph_chunks),
                word_count = self.chunk_size,
                document_id = str(self.id),
                chunk_id = str(uuid5(NAMESPACE_OID, f"{str(self.id)}-{self.chunk_index}")),
                chunk_index = self.chunk_index,
                cut_type = self.paragraph_chunks[len(self.paragraph_chunks) - 1]["cut_type"],
            )



from typing import Optional
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.root_dir import get_absolute_path
from cognee.shared.data_models import MonitoringTool

class BaseConfig(BaseSettings):
    data_root_directory: str = get_absolute_path(".data_storage")
    monitoring_tool: object = MonitoringTool.LANGFUSE
    graphistry_username: Optional[str] = None
    graphistry_password: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "data_root_directory": self.data_root_directory,
            "monitoring_tool": self.monitoring_tool,
        }

@lru_cache
def get_base_config():
    return BaseConfig()



from sqlalchemy.future import select
from cognee.infrastructure.databases.relational import get_relational_engine
from ...models import User, ACL, Resource, Permission

async def give_permission_on_document(
    user: User,
    document_id: str,
    permission_name: str,
):
    db_engine = get_relational_engine()

    document_resource = Resource(resource_id = document_id)

    async with db_engine.get_async_session() as session:
        permission = (await session.execute(select(Permission).filter(Permission.name == permission_name))).scalars().first()

        if permission is None:
            permission = Permission(name = permission_name)

        acl = ACL(principal_id = user.id)
        acl.permission = permission
        acl.resources.append(document_resource)

        session.add(acl)

        await session.commit()



from typing import Dict

class VectorConfig(Dict):
    vector_db_url: str
    vector_db_key: str
    vector_db_provider: str

def create_vector_engine(config: VectorConfig, embedding_engine):
    if config["vector_db_provider"] == "weaviate":
        from .weaviate_db import WeaviateAdapter

        if config["vector_db_url"] is None and config["vector_db_key"] is None:
            raise EnvironmentError("Weaviate is not configured!")

        return WeaviateAdapter(
            config["vector_db_url"],
            config["vector_db_key"],
            embedding_engine = embedding_engine
        )
    elif config["vector_db_provider"] == "qdrant":
        if config["vector_db_url"] and config["vector_db_key"]:
            from .qdrant.QDrantAdapter import QDrantAdapter

            return QDrantAdapter(
                url = config["vector_db_url"],
                api_key = config["vector_db_key"],
                embedding_engine = embedding_engine
            )
    else:
        from .lancedb.LanceDBAdapter import LanceDBAdapter

        return LanceDBAdapter(
            url = config["vector_db_url"],
            api_key = config["vector_db_key"],
            embedding_engine = embedding_engine,
        )

    raise EnvironmentError(f"Vector provider not configured correctly: {config['vector_db_provider']}")



from datetime import datetime, timezone
from sqlalchemy import Column, DateTime, UUID, ForeignKey
from cognee.infrastructure.databases.relational import Base

class DatasetData(Base):
    __tablename__ = "dataset_data"

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))

    dataset_id = Column(UUID(as_uuid = True), ForeignKey("datasets.id"), primary_key = True)
    data_id = Column(UUID(as_uuid = True), ForeignKey("data.id"), primary_key = True)



from uuid import UUID
from sqlalchemy import select
from cognee.modules.data.models import Data, Dataset
from cognee.infrastructure.databases.relational import get_relational_engine

async def get_dataset_data(dataset_id: UUID = None, dataset_name: str = None):
    if dataset_id is None and dataset_name is None:
        raise ValueError("get_dataset_data: Either dataset_id or dataset_name must be provided.")

    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        result = await session.execute(
            select(Data).join(Data.datasets).filter((Dataset.id == dataset_id) | (Dataset.name == dataset_name))
        )
        data = result.scalars().all()

        return data



import logging
from contextlib import asynccontextmanager
from sqlalchemy.ext.asyncio import async_scoped_session
logger = logging.getLogger(__name__)

@asynccontextmanager
async def with_rollback(session: async_scoped_session):
    """Provide a transactional scope around a series of operations."""

    try:
        # async with session.begin():
        yield session
        await session.commit()
        await session.remove()
    except Exception as exception:
        await session.rollback()
        logger.error("Session rolled back due to: %s", str(exception))
        raise exception




from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk


# from cognee.infrastructure.databases.vector import get_vector_engine


async def chunk_remove_disconnected(data_chunks: list[DocumentChunk]) -> list[DocumentChunk]:
    graph_engine = await get_graph_engine()

    document_ids = set((data_chunk.document_id for data_chunk in data_chunks))

    obsolete_chunk_ids = []

    for document_id in document_ids:
        chunk_ids = await graph_engine.get_successor_ids(document_id, edge_label = "has_chunk")

        for chunk_id in chunk_ids:
            previous_chunks = await graph_engine.get_predecessor_ids(chunk_id, edge_label = "next_chunk")

            if len(previous_chunks) == 0:
                obsolete_chunk_ids.append(chunk_id)

    if len(obsolete_chunk_ids) > 0:
        await graph_engine.delete_nodes(obsolete_chunk_ids)

    disconnected_nodes = await graph_engine.get_disconnected_nodes()
    if len(disconnected_nodes) > 0:
        await graph_engine.delete_nodes(disconnected_nodes)

    return data_chunks



from abc import abstractmethod
from typing import Protocol, TypeVar, Type, List

RowDataType = TypeVar('RowDataType')

class RelationalDBInterface(Protocol):
    @abstractmethod
    async def create_database(self, database_name: str, database_path: str): raise NotImplementedError

    @abstractmethod
    async def create_table(self, table_name: str, table_config: object): raise NotImplementedError

    @abstractmethod
    async def add_row(self, table_name: str, row_data: Type[RowDataType]): raise NotImplementedError

    @abstractmethod
    async def add_rows(self, table_name: str, rows_data: List[Type[RowDataType]]): raise NotImplementedError

    @abstractmethod
    async def get_row(self, table_name: str, row_id: str): raise NotImplementedError

    @abstractmethod
    async def update_row(self, table_name: str, row_id: str, row_data: Type[RowDataType]): raise NotImplementedError

    @abstractmethod
    async def delete_row(self, table_name: str, row_id: str): raise NotImplementedError



from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict

class EmbeddingConfig(BaseSettings):
    openai_embedding_model: str = "text-embedding-3-large"
    openai_embedding_dimensions: int = 3072
    litellm_embedding_model: str = "BAAI/bge-large-en-v1.5"
    litellm_embedding_dimensions: int = 1024
    # embedding_engine:object = DefaultEmbeddingEngine(embedding_model=litellm_embedding_model, embedding_dimensions=litellm_embedding_dimensions)

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "openai_embedding_model": self.openai_embedding_model,
            "openai_embedding_dimensions": self.openai_embedding_dimensions,
            "litellm_embedding_model": self.litellm_embedding_model,
            "litellm_embedding_dimensions": self.litellm_embedding_dimensions,
        }

@lru_cache
def get_embedding_config():
    return EmbeddingConfig()



from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.infrastructure.llm.prompts import render_prompt

async def extract_content_graph(content: str, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = render_prompt("generate_graph_prompt.txt", {})
    content_graph = await llm_client.acreate_structured_output(content, system_prompt, response_model)

    return content_graph



from typing import Generic, TypeVar
from pydantic import BaseModel

PayloadSchema = TypeVar("PayloadSchema", bound = BaseModel)

class DataPoint(BaseModel, Generic[PayloadSchema]):
    id: str
    payload: PayloadSchema
    embed_field: str = "value"

    def get_embeddable_data(self):
        if hasattr(self.payload, self.embed_field):
            return getattr(self.payload, self.embed_field)



from uuid import uuid4
from datetime import datetime, timezone
# from sqlalchemy.orm import relationship
from sqlalchemy import Column, DateTime, UUID, String
from cognee.infrastructure.databases.relational import Base

class Permission(Base):
    __tablename__ = "permissions"

    id = Column(UUID, primary_key = True, index = True, default = uuid4)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))
    updated_at = Column(DateTime(timezone = True), onupdate = lambda: datetime.now(timezone.utc))

    name = Column(String, unique = True, nullable = False, index = True)

    # acls = relationship("ACL", back_populates = "permission")



from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import Column, String, ForeignKey, UUID
from .Principal import Principal
from .UserGroup import UserGroup

class Group(Principal):
    __tablename__ = "groups"

    id = Column(UUID(as_uuid = True), ForeignKey("principals.id"), primary_key = True)

    name = Column(String, unique = True, nullable = False, index = True)

    users: Mapped[list["User"]] = relationship(
        "User",
        secondary = UserGroup.__tablename__,
        back_populates = "groups",
    )

    __mapper_args__ = {
        "polymorphic_identity": "group",
    }



from uuid import UUID
from sqlalchemy import func, select
from sqlalchemy.orm import aliased
from cognee.infrastructure.databases.relational import get_relational_engine
from ..models import PipelineRun

async def get_pipeline_status(pipeline_ids: list[UUID]):
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        query = select(
            PipelineRun,
            func.row_number().over(
                partition_by = PipelineRun.run_id,
                order_by = PipelineRun.created_at.desc(),
            ).label("rn")
        ).filter(PipelineRun.run_id.in_(pipeline_ids)).subquery()

        aliased_pipeline_run = aliased(PipelineRun, query)

        latest_runs = (
            select(aliased_pipeline_run).filter(query.c.rn == 1)
        )
      
        runs = (await session.execute(latest_runs)).scalars().all()

        pipeline_statuses = {
            str(run.run_id): run.status for run in runs
        }

        return pipeline_statuses

        # f"""SELECT data_id, status
        # FROM (
        #     SELECT data_id, status, ROW_NUMBER() OVER (PARTITION BY data_id ORDER BY created_at DESC) as rn
        #     FROM cognee.cognee.task_runs
        #     WHERE data_id IN ({formatted_data_ids})
        # ) t
        # WHERE rn = 1;"""

      # return { dataset["data_id"]: dataset["status"] for dataset in datasets_statuses }



from typing import Type, Dict
from pydantic import BaseModel
from cognee.infrastructure.llm.prompts import render_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client

async def extract_cognitive_layers(content: str, category: Dict, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = render_prompt("generate_cog_layers.txt", category)

    return await llm_client.acreate_structured_output(content, system_prompt, response_model)



import asyncio
from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.infrastructure.databases.vector import get_vector_engine

async def search_adjacent(query: str) -> list[(str, str)]:
    """
    Find the neighbours of a given node in the graph and return their ids and descriptions.

    Parameters:
    - query (str): The query string to filter nodes by.

    Returns:
    - list[(str, str)]: A list containing the unique identifiers and names of the neighbours of the given node.
    """
    node_id = query

    if node_id is None:
        return {}

    graph_engine = await get_graph_engine()

    exact_node = await graph_engine.extract_node(node_id)

    if exact_node is not None and "uuid" in exact_node:
        neighbours = await graph_engine.get_neighbours(exact_node["uuid"])
    else:
        vector_engine = get_vector_engine()
        results = await asyncio.gather(
            vector_engine.search("entities", query_text = query, limit = 10),
            vector_engine.search("classification", query_text = query, limit = 10),
        )
        results = [*results[0], *results[1]]
        relevant_results = [result for result in results if result.score < 0.5][:5]

        if len(relevant_results) == 0:
            return []

        node_neighbours = await asyncio.gather(*[graph_engine.get_neighbours(result.id) for result in relevant_results])
        neighbours = []
        for neighbour_ids in node_neighbours:
            neighbours.extend(neighbour_ids)

    return neighbours



""" FalcorDB Adapter for Graph Database"""
import json
import logging
from typing import Optional, Any, List, Dict
from contextlib import asynccontextmanager


from falkordb.asyncio import FalkorDB
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

logger = logging.getLogger("FalcorDBAdapter")

class FalcorDBAdapter(GraphDBInterface):
    def __init__(
        self,
        graph_database_url: str,
        graph_database_username: str,
        graph_database_password: str,
        graph_database_port: int,
        driver: Optional[Any] = None,
        graph_name: str = "DefaultGraph",
    ):
        self.driver = FalkorDB(
            host = graph_database_url,
            port = graph_database_port)
        self.graph_name = graph_name



    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        try:
                selected_graph = self.driver.select_graph(self.graph_name)

                result = await selected_graph.query(query)
                return result.result_set

        except Exception as error:
            logger.error("Falkor query error: %s", error, exc_info = True)
            raise error

    async def graph(self):
        return  self.driver

    async def add_node(self, node_id: str, node_properties: Dict[str, Any] = None):
        node_id = node_id.replace(":", "_")

        serialized_properties = self.serialize_properties(node_properties)

        if "name" not in serialized_properties:
            serialized_properties["name"] = node_id

        # serialized_properties["created_at"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        # serialized_properties["updated_at"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # properties = ", ".join(f"{property_name}: ${property_name}" for property_name in serialized_properties.keys())

        query = f"""MERGE (node:`{node_id}` {{id: $node_id}})
                ON CREATE SET node += $properties
                RETURN ID(node) AS internal_id, node.id AS nodeId"""

        params = {
            "node_id": node_id,
            "properties": serialized_properties,
        }

        return await self.query(query, params)

    async def add_nodes(self, nodes: list[tuple[str, dict[str, Any]]]) -> None:
        for node in nodes:
            node_id, node_properties = node
            node_id = node_id.replace(":", "_")

            await self.add_node(
                node_id = node_id,
                node_properties = node_properties,
            )



    async def extract_node_description(self, node_id: str):
        query = """MATCH (n)-[r]->(m)
                    WHERE n.id = $node_id
                    AND NOT m.id CONTAINS 'DefaultGraphModel'
                    RETURN m
                    """

        result = await self.query(query, dict(node_id = node_id))

        descriptions = []

        for node in result:
            # Assuming 'm' is a consistent key in your data structure
            attributes = node.get("m", {})

            # Ensure all required attributes are present
            if all(key in attributes for key in ["id", "layer_id", "description"]):
                descriptions.append({
                    "id": attributes["id"],
                    "layer_id": attributes["layer_id"],
                    "description": attributes["description"],
                })

        return descriptions

    async def get_layer_nodes(self):
        query = """MATCH (node) WHERE node.layer_id IS NOT NULL
        RETURN node"""

        return [result["node"] for result in (await self.query(query))]

    async def extract_node(self, node_id: str):
        results = self.extract_nodes([node_id])

        return results[0] if len(results) > 0 else None

    async def extract_nodes(self, node_ids: List[str]):
        query = """
        UNWIND $node_ids AS id
        MATCH (node {id: id})
        RETURN node"""

        params = {
            "node_ids": node_ids
        }

        results = await self.query(query, params)

        return results

    async def delete_node(self, node_id: str):
        node_id = id.replace(":", "_")

        query = f"MATCH (node:`{node_id}` {{id: $node_id}}) DETACH DELETE n"
        params = { "node_id": node_id }

        return await self.query(query, params)

    async def add_edge(self, from_node: str, to_node: str, relationship_name: str, edge_properties: Optional[Dict[str, Any]] = {}):
        serialized_properties = self.serialize_properties(edge_properties)
        from_node = from_node.replace(":", "_")
        to_node = to_node.replace(":", "_")

        query = f"""MATCH (from_node:`{from_node}` {{id: $from_node}}), (to_node:`{to_node}` {{id: $to_node}})
                MERGE (from_node)-[r:`{relationship_name}`]->(to_node)
                SET r += $properties
                RETURN r"""

        params = {
            "from_node": from_node,
            "to_node": to_node,
            "properties": serialized_properties
        }

        return await self.query(query, params)


    async def add_edges(self, edges: list[tuple[str, str, str, dict[str, Any]]]) -> None:
        # edges_data = []

        for edge in edges:
            from_node, to_node, relationship_name, edge_properties = edge
            from_node = from_node.replace(":", "_")
            to_node = to_node.replace(":", "_")

            await self.add_edge(
                from_node = from_node,
                to_node = to_node,
                relationship_name = relationship_name,
                edge_properties = edge_properties
            )



    async def filter_nodes(self, search_criteria):
        query = f"""MATCH (node)
                WHERE node.id CONTAINS '{search_criteria}'
                RETURN node"""


        return await self.query(query)


    async def delete_graph(self):
        query = """MATCH (node)
                DETACH DELETE node;"""

        return await self.query(query)

    def serialize_properties(self, properties = dict()):
        return {
            property_key: json.dumps(property_value)
            if isinstance(property_value, (dict, list))
            else property_value for property_key, property_value in properties.items()
        }



from typing import Type
from pydantic import BaseModel
import instructor
from tenacity import retry, stop_after_attempt
import anthropic
from cognee.infrastructure.llm.llm_interface import LLMInterface
from cognee.infrastructure.llm.prompts import read_query_prompt


class AnthropicAdapter(LLMInterface):
    """Adapter for Anthropic API"""
    name = "Anthropic"
    model: str

    def __init__(self, model: str = None):
        self.aclient = instructor.patch(
            create = anthropic.Anthropic().messages.create,
            mode = instructor.Mode.ANTHROPIC_TOOLS
        )
        self.model = model

    async def acreate_structured_output(
        self,
        text_input: str,
        system_prompt: str,
        response_model: Type[BaseModel]
    ) -> BaseModel:
        """Generate a response from a user query."""

        return await self.aclient(
            model = self.model,
            max_tokens = 4096,
            max_retries = 5,
            messages = [{
                "role": "user",
                "content": f"""Use the given format to extract information
                from the following input: {text_input}. {system_prompt}""",
            }],
            response_model = response_model,
        )

    def show_prompt(self, text_input: str, system_prompt: str) -> str:
        """Format and display the prompt for a user query."""

        if not text_input:
            text_input = "No user input provided."
        if not system_prompt:
            raise ValueError("No system prompt path provided.")

        system_prompt = read_query_prompt(system_prompt)

        formatted_prompt = f"""System Prompt:\n{system_prompt}\n\nUser Input:\n{text_input}\n""" if system_prompt else None

        return formatted_prompt



from typing import Protocol

class EmbeddingEngine(Protocol):
    async def embed_text(self, text: list[str]) -> list[list[float]]:
        raise NotImplementedError()

    def get_vector_size(self) -> int:
        raise NotImplementedError()



import os
import asyncio
from typing import AsyncGenerator
from contextlib import asynccontextmanager
from sqlalchemy import create_engine, text, select
from sqlalchemy.orm import sessionmaker, joinedload
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, async_sessionmaker
from cognee.infrastructure.files.storage import LocalStorage
from cognee.infrastructure.databases.relational.FakeAsyncSession import FakeAsyncSession
from ..ModelBase import Base

def make_async_sessionmaker(sessionmaker):
    @asynccontextmanager
    async def async_session_maker():
        await asyncio.sleep(0.1)
        session = FakeAsyncSession(sessionmaker())
        try:
            yield session
        finally:
            await session.close()  # Ensure the session is closed

    return async_session_maker

class SQLAlchemyAdapter():
    def __init__(self, db_type: str, db_path: str, db_name: str, db_user: str, db_password: str, db_host: str, db_port: str):
        self.db_location = os.path.abspath(os.path.join(db_path, db_name))
        self.db_name = db_name

        if db_type == "duckdb":
            LocalStorage.ensure_directory_exists(db_path)

            self.engine = create_engine(f"duckdb:///{self.db_location}")
            self.sessionmaker = make_async_sessionmaker(sessionmaker(bind=self.engine))
        else:
            self.engine = create_async_engine(f"postgresql+asyncpg://{db_user}:{db_password}@{db_host}:{db_port}/{db_name}")
            self.sessionmaker = async_sessionmaker(bind=self.engine, expire_on_commit=False)

    @asynccontextmanager
    async def get_async_session(self) -> AsyncGenerator[AsyncSession, None]:
        async_session_maker = self.sessionmaker
        async with async_session_maker() as session:
            try:
                yield session
            finally:
                await session.close()  # Ensure the session is closed

    def get_session(self):
        session_maker = self.sessionmaker
        with session_maker() as session:
            try:
                yield session
            finally:
                session.close()  # Ensure the session is closed

    async def get_datasets(self):
        from cognee.modules.data.models import Dataset

        async with self.get_async_session() as session:
            result = await session.execute(select(Dataset).options(joinedload(Dataset.data)))
            datasets = result.unique().scalars().all()
            return datasets

    async def create_table(self, schema_name: str, table_name: str, table_config: list[dict]):
        fields_query_parts = [f"{item['name']} {item['type']}" for item in table_config]
        async with self.engine.begin() as connection:
            await connection.execute(text(f"CREATE SCHEMA IF NOT EXISTS {schema_name};"))
            await connection.execute(text(f"CREATE TABLE IF NOT EXISTS {schema_name}.{table_name} ({', '.join(fields_query_parts)});"))
            await connection.close()

    async def delete_table(self, table_name: str):
        async with self.engine.begin() as connection:
            await connection.execute(text(f"DROP TABLE IF EXISTS {table_name} CASCADE;"))

            await connection.close()
    async def insert_data(self, schema_name: str, table_name: str, data: list[dict]):
        columns = ", ".join(data[0].keys())
        values = ", ".join([f"({', '.join([f':{key}' for key in row.keys()])})" for row in data])
        insert_query = text(f"INSERT INTO {schema_name}.{table_name} ({columns}) VALUES {values};")

        async with self.engine.begin() as connection:
            await connection.execute(insert_query, data)
            await connection.close()
    async def get_data(self, table_name: str, filters: dict = None):
        async with self.engine.begin() as connection:
            query = f"SELECT * FROM {table_name}"
            if filters:
                filter_conditions = " AND ".join([
                    f"{key} IN ({', '.join([f':{key}{i}' for i in range(len(value))])})" if isinstance(value, list)
                    else f"{key} = :{key}" for key, value in filters.items()
                ])
                query += f" WHERE {filter_conditions};"
                query = text(query)
                results = await connection.execute(query, filters)
            else:
                query += ";"
                query = text(query)
                results = await connection.execute(query)
            return {result["data_id"]: result["status"] for result in results}

    async def execute_query(self, query):
        async with self.engine.begin() as connection:
            result = await connection.execute(text(query))
            return [dict(row) for row in result]

    async def drop_tables(self):
        async with self.engine.begin() as connection:
            try:
                await connection.execute(text("DROP TABLE IF EXISTS group_permission CASCADE"))
                await connection.execute(text("DROP TABLE IF EXISTS permissions CASCADE"))
                # Add more DROP TABLE statements for other tables as needed
                print("Database tables dropped successfully.")
            except Exception as e:
                print(f"Error dropping database tables: {e}")

    async def delete_database(self):
        async with self.engine.begin() as connection:
            try:
                for table in Base.metadata.sorted_tables:
                    drop_table_query = text(f'DROP TABLE IF EXISTS {table.name} CASCADE')
                    await connection.execute(drop_table_query)
                print("Database deleted successfully.")
            except Exception as e:
                print(f"Error deleting database: {e}")



from cognee.base_config import get_base_config
from cognee.infrastructure.files.storage import LocalStorage

async def prune_data():
    base_config = get_base_config()
    data_root_directory = base_config.data_root_directory
    LocalStorage.remove_all(data_root_directory)



from pydantic import BaseModel
from cognee.infrastructure.llm import get_llm_config

class LLMConfig(BaseModel):
    apiKey: str
    model: str
    provider: str

async def save_llm_config(new_llm_config: LLMConfig):
    llm_config = get_llm_config()

    llm_config.llm_provider = new_llm_config.provider
    llm_config.llm_model = new_llm_config.model

    if "*****" not in new_llm_config.apiKey and len(new_llm_config.apiKey.strip()) > 0:
        llm_config.llm_api_key = new_llm_config.apiKey



from datetime import datetime, timezone
from sqlalchemy import Column, ForeignKey, DateTime, UUID
from cognee.infrastructure.databases.relational import Base

class UserGroup(Base):
    __tablename__ = "user_groups"

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))

    user_id = Column(UUID(as_uuid = True), ForeignKey("users.id"), primary_key = True)
    group_id = Column(UUID(as_uuid = True), ForeignKey("groups.id"), primary_key = True)



from cognee.modules.users.get_fastapi_users import get_fastapi_users
from cognee.modules.users.models.User import UserRead, UserUpdate

def get_users_router():
    return get_fastapi_users().get_users_router(UserRead, UserUpdate)



chunking_registry = {}

def register_chunking_function(name):
    def decorator(func):
        chunking_registry[name] = func
        return func
    return decorator

def get_chunking_function(name: str):
    return chunking_registry.get(name)



from typing import List, Dict, Optional, Any

from falkordb import FalkorDB
from qdrant_client import AsyncQdrantClient, models
from ..vector_db_interface import VectorDBInterface
from ..models.DataPoint import DataPoint
from ..embeddings.EmbeddingEngine import EmbeddingEngine




class FalcorDBAdapter(VectorDBInterface):
    def __init__(
        self,
        graph_database_url: str,
        graph_database_username: str,
        graph_database_password: str,
        graph_database_port: int,
        driver: Optional[Any] = None,
        embedding_engine =  EmbeddingEngine,
        graph_name: str = "DefaultGraph",
    ):
        self.driver = FalkorDB(
            host = graph_database_url,
            port = graph_database_port)
        self.graph_name = graph_name
        self.embedding_engine = embedding_engine



    async def embed_data(self, data: list[str]) -> list[list[float]]:
        return await self.embedding_engine.embed_text(data)


    async def create_collection(self, collection_name: str, payload_schema = None):
        pass


    async def create_data_points(self, collection_name: str, data_points: List[DataPoint]):
        pass

    async def retrieve(self, collection_name: str, data_point_ids: list[str]):
        pass

    async def search(
        self,
        collection_name: str,
        query_text: str = None,
        query_vector: List[float] = None,
        limit: int = 10,
        with_vector: bool = False,
    ):
        pass

    async def delete_data_points(self, collection_name: str, data_point_ids: list[str]):
        pass



from cognee.infrastructure.databases.vector import DataPoint, get_vector_engine
from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk

async def save_chunks_to_store(data_chunks: list[DocumentChunk], collection_name: str):
    if len(data_chunks) == 0:
        return data_chunks

    vector_engine = get_vector_engine()
    graph_engine = await get_graph_engine()

    # Remove and unlink existing chunks
    if await vector_engine.has_collection(collection_name):
        existing_chunks = [DocumentChunk.parse_obj(chunk.payload) for chunk in (await vector_engine.retrieve(
            collection_name,
            [str(chunk.chunk_id) for chunk in data_chunks],
        ))]

        if len(existing_chunks) > 0:
            await vector_engine.delete_data_points(collection_name, [str(chunk.chunk_id) for chunk in existing_chunks])

            await graph_engine.remove_connection_to_successors_of([chunk.chunk_id for chunk in existing_chunks], "next_chunk")
            await graph_engine.remove_connection_to_predecessors_of([chunk.chunk_id for chunk in existing_chunks], "has_chunk")
    else:
        await vector_engine.create_collection(collection_name, payload_schema = DocumentChunk)

    # Add to vector storage
    await vector_engine.create_data_points(
        collection_name,
        [
            DataPoint[DocumentChunk](
                id = str(chunk.chunk_id),
                payload = chunk,
                embed_field = "text",
            ) for chunk in data_chunks
        ],
    )

    # Add to graph storage
    chunk_nodes = []
    chunk_edges = []

    for chunk in data_chunks:
        chunk_nodes.append((
            str(chunk.chunk_id),
            dict(
                id = str(chunk.chunk_id),
                chunk_id = str(chunk.chunk_id),
                document_id = str(chunk.document_id),
                word_count = chunk.word_count,
                chunk_index = chunk.chunk_index,
                cut_type = chunk.cut_type,
            )
        ))

        chunk_edges.append((
            str(chunk.document_id),
            str(chunk.chunk_id),
            "has_chunk",
            dict(
                relationship_name = "has_chunk",
                source_node_id = str(chunk.document_id),
                target_node_id = str(chunk.chunk_id),
            ),
        ))

        previous_chunk_id = get_previous_chunk_id(data_chunks, chunk)

        if previous_chunk_id is not None:
            chunk_edges.append((
                str(previous_chunk_id),
                str(chunk.chunk_id),
                "next_chunk",
                dict(
                    relationship_name = "next_chunk",
                    source_node_id = str(previous_chunk_id),
                    target_node_id = str(chunk.chunk_id),
                ),
            ))

    await graph_engine.add_nodes(chunk_nodes)
    await graph_engine.add_edges(chunk_edges)

    return data_chunks


def get_previous_chunk_id(document_chunks: list[DocumentChunk], current_chunk: DocumentChunk) -> DocumentChunk:
    if current_chunk.chunk_index == 0:
        return current_chunk.document_id

    for chunk in document_chunks:
        if str(chunk.document_id) == str(current_chunk.document_id) \
            and chunk.chunk_index == current_chunk.chunk_index - 1:
            return chunk.chunk_id

    return None



import asyncio
from uuid import uuid5, NAMESPACE_OID
from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.infrastructure.databases.vector import get_vector_engine, DataPoint
from cognee.modules.data.extraction.extract_categories import extract_categories
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk


async def chunk_naive_llm_classifier(data_chunks: list[DocumentChunk], classification_model: Type[BaseModel]):
    if len(data_chunks) == 0:
        return data_chunks

    chunk_classifications = await asyncio.gather(
        *[extract_categories(chunk.text, classification_model) for chunk in data_chunks],
    )

    classification_data_points = []

    for chunk_index, chunk in enumerate(data_chunks):
        chunk_classification = chunk_classifications[chunk_index]
        classification_data_points.append(uuid5(NAMESPACE_OID, chunk_classification.label.type))
        classification_data_points.append(uuid5(NAMESPACE_OID, chunk_classification.label.type))

        for classification_subclass in chunk_classification.label.subclass:
            classification_data_points.append(uuid5(NAMESPACE_OID, classification_subclass.value))

    vector_engine = get_vector_engine()

    class Keyword(BaseModel):
        uuid: str
        text: str
        chunk_id: str
        document_id: str

    collection_name = "classification"

    if await vector_engine.has_collection(collection_name):
        existing_data_points = await vector_engine.retrieve(
            collection_name,
            list(set(classification_data_points)),
        ) if len(classification_data_points) > 0 else []

        existing_points_map = {point.id: True for point in existing_data_points}
    else:
        existing_points_map = {}
        await vector_engine.create_collection(collection_name, payload_schema=Keyword)

    data_points = []
    nodes = []
    edges = []

    for (chunk_index, data_chunk) in enumerate(data_chunks):
        chunk_classification = chunk_classifications[chunk_index]
        classification_type_label = chunk_classification.label.type
        classification_type_id = uuid5(NAMESPACE_OID, classification_type_label)

        if classification_type_id not in existing_points_map:
            data_points.append(
                DataPoint[Keyword](
                    id=str(classification_type_id),
                    payload=Keyword.parse_obj({
                        "uuid": str(classification_type_id),
                        "text": classification_type_label,
                        "chunk_id": str(data_chunk.chunk_id),
                        "document_id": str(data_chunk.document_id),
                    }),
                    embed_field="text",
                )
            )

            nodes.append((
                str(classification_type_id),
                dict(
                    id=str(classification_type_id),
                    name=classification_type_label,
                    type=classification_type_label,
                )
            ))
            existing_points_map[classification_type_id] = True

        edges.append((
            str(data_chunk.chunk_id),
            str(classification_type_id),
            "is_media_type",
            dict(
                relationship_name="is_media_type",
                source_node_id=str(data_chunk.chunk_id),
                target_node_id=str(classification_type_id),
            ),
        ))

        for classification_subclass in chunk_classification.label.subclass:
            classification_subtype_label = classification_subclass.value
            classification_subtype_id = uuid5(NAMESPACE_OID, classification_subtype_label)

            if classification_subtype_id not in existing_points_map:
                data_points.append(
                    DataPoint[Keyword](
                        id=str(classification_subtype_id),
                        payload=Keyword.parse_obj({
                            "uuid": str(classification_subtype_id),
                            "text": classification_subtype_label,
                            "chunk_id": str(data_chunk.chunk_id),
                            "document_id": str(data_chunk.document_id),
                        }),
                        embed_field="text",
                    )
                )

                nodes.append((
                    str(classification_subtype_id),
                    dict(
                        id=str(classification_subtype_id),
                        name=classification_subtype_label,
                        type=classification_subtype_label,
                    )
                ))
                edges.append((
                    str(classification_subtype_id),
                    str(classification_type_id),
                    "is_subtype_of",
                    dict(
                        relationship_name="contains",
                        source_node_id=str(classification_type_id),
                        target_node_id=str(classification_subtype_id),
                    ),
                ))

                existing_points_map[classification_subtype_id] = True

            edges.append((
                str(data_chunk.chunk_id),
                str(classification_subtype_id),
                "is_classified_as",
                dict(
                    relationship_name="is_classified_as",
                    source_node_id=str(data_chunk.chunk_id),
                    target_node_id=str(classification_subtype_id),
                ),
            ))

    if len(nodes) > 0 or len(edges) > 0:
        await vector_engine.create_data_points(collection_name, data_points)

        graph_engine = await get_graph_engine()

        await graph_engine.add_nodes(nodes)
        await graph_engine.add_edges(edges)

    return data_chunks



import logging
from typing import List, Dict
from cognee.modules.cognify.config import get_cognify_config
from .extraction.categorize_relevant_summary import categorize_relevant_summary

logger = logging.getLogger(__name__)
async def get_cognitive_layers(content: str, categories: List[Dict]):
    try:
        cognify_config = get_cognify_config()
        return (await categorize_relevant_summary(
            content,
            categories[0],
            cognify_config.summarization_model,
        )).cognitive_layers
    except Exception as error:
        logger.error("Error extracting cognitive layers from content: %s", error, exc_info = True)
        raise error



import asyncio
from cognee.modules.pipelines.operations.run_tasks import run_tasks
from cognee.modules.pipelines.tasks.Task import Task


async def main():
    def number_generator(num):
        for i in range(num):
            yield i + 1

    async def add_one(nums):
        for num in nums:
            yield num + 1

    async def multiply_by_two(num):
        yield num * 2

    async def add_one_single(num):
        yield num + 1

    pipeline = run_tasks([
        Task(number_generator),
        Task(add_one, task_config = {"batch_size": 5}),
        Task(multiply_by_two, task_config = {"batch_size": 1}),
        Task(add_one_single),
    ], 10)

    results = [5, 7, 9, 11, 13, 15, 17, 19, 21, 23]
    index = 0
    async for result in pipeline:
        print(result)
        assert result == results[index]
        index += 1

if __name__ == "__main__":
    asyncio.run(main())



from io import BufferedReader
from typing import Union, BinaryIO
from .exceptions import IngestionException
from .data_types import TextData, BinaryData

def classify(data: Union[str, BinaryIO], filename: str = None):
    if isinstance(data, str):
        return TextData(data)

    if isinstance(data, BufferedReader):
        return BinaryData(data, data.name.split("/")[-1] if data.name else filename)

    if hasattr(data, "file"):
        return BinaryData(data.file, filename)

    raise IngestionException(f"Type of data sent to classify(data: Union[str, BinaryIO) not supported: {type(data)}")



import dspy
from cognee.modules.data.extraction.knowledge_graph.extract_knowledge_graph_module import ExtractKnowledgeGraph
from cognee.root_dir import get_absolute_path
from cognee.infrastructure.llm import get_llm_config

def run():
    llm_config = get_llm_config()
    gpt4 = dspy.OpenAI(model = llm_config.llm_model, api_key = llm_config.llm_api_key, model_type = "chat", max_tokens = 4096)
    compiled_extract_knowledge_graph = ExtractKnowledgeGraph(lm = gpt4)
    compiled_extract_knowledge_graph.load(get_absolute_path("./programs/extract_knowledge_graph/extract_knowledge_graph.json"))

    text = """The 1985 FA Charity Shield (also known as the General Motors FA
              Charity Shield for sponsorship reasons) was the 63rd FA Charity Shield,
              an annual football match played between the winners of the previous
              season's First Division and FA Cup competitions. The match was played on
              10 August 1985 at Wembley Stadium and contested by Everton,
              who had won the 1984\u201385 First Division, and Manchester United,
              who had won the 1984\u201385 FA Cup. Everton won 2\u20130 with goals from
              Trevor Steven and Adrian Heath. Trevor Steven put Everton into the lead
              when he swept home from six yards after a cross from the left in the first half.
              The second goal came in the second half when Manchester United goalkeeper
              Gary Bailey dropped a cross from the left to allow Adrian Heath to tip the
              ball past him into the left corner of the net.\r\nThe 1995 FA Charity Shield
              (also known as the Littlewoods FA Charity Shield for sponsorship reasons) was the
              73rd FA Charity Shield, an annual football match played between the winners of
              the previous season's Premier League and FA Cup competitions. The match was
              played on 13 August 1995 at Wembley Stadium and contested by Blackburn Rovers,
              who had won the Premier League and FA Cup winners Everton. It was Blackburn's
              second successive Charity Shield appearance, while Everton were appearing in
              their eleventh and their first since 1987. Everton won the match 1\u20130
              with a goal from Vinny Samways when he caught Tim Flowers off his line and
              lifted the ball over him from the left of the penalty area and into the right
              corner of the net. Dave Watson lifted the trophy for Everton.\r\nThe 1972 FA
              Charity Shield was contested between Manchester City and Aston Villa.\r\nThe
              1997 FA Charity Shield (known as the Littlewoods FA Charity Shield for
              sponsorship reasons) was the 75th FA Charity Shield, an annual football match
              played between the winners of the previous season's Premier League and
              FA Cup competitions. The match was played on 3 August 1997 at Wembley Stadium
              and contested by Manchester United, who had won the 1996\u201397 FA Premier League,
              and Chelsea, who had won the 1996\u201397 FA Cup. Manchester United won the match
              4\u20132 on penalties after the match had finished at 1\u20131 after 90 minutes.
              \r\nThe 1956 FA Charity Shield was the 34th FA Charity Shield, an annual football
              match held between the winners of the previous season's Football League and
              FA Cup competitions. The match was contested by Manchester United, who had won
              the 1955\u201356 Football League, and Manchester City, who had won the
              1955\u201356 FA Cup, at Maine Road, Manchester, on 24 October 1956. Manchester
              United won the match 1\u20130, Dennis Viollet scoring the winning goal.
              Manchester United goalkeeper David Gaskell made his debut for the club during
              the game, taking the place of injured goalkeeper Ray Wood, and, at the age of
              16 years and 19 days, became the youngest player ever to play for the club.
              \r\nThe 1937 FA Charity Shield was the 24th FA Charity Shield, a football match
              between the winners of the previous season's First Division and FA Cup competitions.
              The match was contested by league champions Manchester City and FA Cup winners
              Sunderland, and was played at Maine Road, the home ground of Manchester City.
              Manchester City won the game, 2\u20130.\r\nThe 2000 FA Charity Shield (also known
              as the One 2 One FA Charity Shield for sponsorship reasons) was the
              78th FA Charity Shield, an annual football match played between the winners
              of the previous season's Premier League and FA Cup competitions. The match
              was played between Manchester United, who won the 1999\u20132000 Premier League,
              and Chelsea, who won the 1999\u20132000 FA Cup, and resulted in a 2\u20130 Chelsea win.
              The goals were scored by Jimmy Floyd Hasselbaink and Mario Melchiot. Roy Keane
              was sent off for a challenge on Gustavo Poyet and was the last person to be
              sent off at the old Wembley Stadium.\r\nThe 2001 FA Charity Shield (also known
              as the One 2 One FA Charity Shield for sponsorship reasons) was the 79th FA Charity Shield,
              an annual football match played between the winners of the previous season's
              Premier League and FA Cup. The match was contested between Liverpool, winners of
              the 2000\u201301 FA Cup and Manchester United, who won the 2000\u201301 Premier
              League on 12 August 2001. It was the first Shield match to be held at the
              Millennium Stadium following the closure of Wembley Stadium for reconstruction.
              \r\nAston Villa Football Club ( ; nicknamed Villa, The Villa, The Villans
              and The Lions) is a professional football club in Aston, Birmingham, that plays
              in the Championship, the second level of English football. Founded in 1874,
              they have played at their current home ground, Villa Park, since 1897. Aston Villa
              were one of the founder members of the Football League in 1888 and of the
              Premier League in 1992.\r\nThe 1996 FA Charity Shield (also known as the
              Littlewoods FA Charity Shield for sponsorship reasons) was the 74th FA Charity Shield,
              an annual football match played between the winners of the previous season's Premier
              League and FA Cup competitions. The match was played on 11 August 1996 at Wembley
              Stadium and contested by Manchester United, who had won the Double of Premier League
              and FA Cup in 1995\u201396, and Newcastle United, who had finished as runners-up
              in the Premier League. Manchester United won the match 4\u20130 with goals from
              Eric Cantona, Nicky Butt, David Beckham and Roy Keane."""

    prediction = compiled_extract_knowledge_graph(context = text, question = "")

    print(prediction.graph)

if __name__ == "__main__":
    run()



import asyncio
from queue import Queue
from cognee.modules.pipelines.operations.run_tasks import run_tasks
from cognee.modules.pipelines.tasks.Task import Task

async def pipeline(data_queue):
    async def queue_consumer():
        while not data_queue.is_closed:
            if not data_queue.empty():
                yield data_queue.get()
            else:
                await asyncio.sleep(0.3)

    async def add_one(num):
        yield num + 1

    async def multiply_by_two(num):
        yield num * 2

    tasks_run = run_tasks([
        Task(queue_consumer),
        Task(add_one),
        Task(multiply_by_two),
    ])

    results = [2, 4, 6, 8, 10, 12, 14, 16, 18]
    index = 0
    async for result in tasks_run:
        print(result)
        assert result == results[index]
        index += 1

async def main():
    data_queue = Queue()
    data_queue.is_closed = False

    async def queue_producer():
        for i in range(0, 10):
            data_queue.put(i)
            await asyncio.sleep(0.1)
        data_queue.is_closed = True

    await asyncio.gather(pipeline(data_queue), queue_producer())

if __name__ == "__main__":
    asyncio.run(main())



from typing import BinaryIO, TypedDict
from .guess_file_type import guess_file_type


class FileMetadata(TypedDict):
    name: str
    mime_type: str
    extension: str

def get_file_metadata(file: BinaryIO) -> FileMetadata:
    """Get metadata from a file"""
    file.seek(0)
    file_type = guess_file_type(file)

    file_path = file.name
    file_name = file_path.split("/")[-1].split(".")[0] if file_path else None

    return FileMetadata(
        name = file_name,
        file_path = file_path,
        mime_type = file_type.mime,
        extension = file_type.extension,
    )



import os
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.root_dir import get_absolute_path

class VectorConfig(BaseSettings):
    vector_db_url: str = os.path.join(
        os.path.join(get_absolute_path(".cognee_system"), "databases"),
        "cognee.lancedb"
    )
    vector_db_key: str = ""
    vector_engine_provider: str = "lancedb"

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "vector_db_url": self.vector_db_url,
            "vector_db_key": self.vector_db_key,
            "vector_db_provider": self.vector_engine_provider,
        }

@lru_cache
def get_vectordb_config():
    return VectorConfig()



from cognee.modules.users.get_fastapi_users import get_fastapi_users
from cognee.modules.users.authentication.get_auth_backend import get_auth_backend

def get_auth_router():
    auth_backend = get_auth_backend()
    return get_fastapi_users().get_auth_router(auth_backend)



import asyncio
from cognee.infrastructure.databases.graph.get_graph_engine import get_graph_engine
from cognee.infrastructure.databases.vector import get_vector_engine

async def search_traverse(query: str):
    node_id = query
    rules = set()

    graph_engine = await get_graph_engine()
    vector_engine = get_vector_engine()

    exact_node = await graph_engine.extract_node(node_id)

    if exact_node is not None and "uuid" in exact_node:
        edges = await graph_engine.get_edges(exact_node["uuid"])

        for edge in edges:
            rules.add(f"{edge[0]} {edge[2]['relationship_name']} {edge[1]}")
    else:
        results = await asyncio.gather(
            vector_engine.search("entities", query_text = query, limit = 10),
            vector_engine.search("classification", query_text = query, limit = 10),
        )
        results = [*results[0], *results[1]]
        relevant_results = [result for result in results if result.score < 0.5][:5]

        if len(relevant_results) > 0:
            for result in relevant_results:
                graph_node_id = result.id

                edges = await graph_engine.get_edges(graph_node_id)

                for edge in edges:
                    rules.add(f"{edge[0]} {edge[2]['relationship_name']} {edge[1]}")

    return list(rules)



from typing import Union, Literal
from pydantic import BaseModel
from cognee.infrastructure.databases.vector import get_vectordb_config

class VectorDBConfig(BaseModel):
    url: str
    apiKey: str
    provider: Union[Literal["lancedb"], Literal["qdrant"], Literal["weaviate"]]

async def save_vector_db_config(vector_db_config: VectorDBConfig):
    vector_config = get_vectordb_config()

    vector_config.vector_db_url = vector_db_config.url
    vector_config.vector_db_key = vector_db_config.apiKey
    vector_config.vector_engine_provider = vector_db_config.provider



from uuid import UUID, uuid5, NAMESPACE_OID
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.modules.data.chunking.TextChunker import TextChunker
from .Document import Document

class AudioDocument(Document):
    type: str = "audio"
    title: str
    file_path: str
    chunking_strategy:str

    def __init__(self, id: UUID, title: str, file_path: str, chunking_strategy:str="paragraph"):
        self.id = id or uuid5(NAMESPACE_OID, title)
        self.title = title
        self.file_path = file_path
        self.chunking_strategy = chunking_strategy

    def read(self):
        # Transcribe the audio file
        result = get_llm_client().create_transcript(self.file_path)
        text = result.text

        chunker = TextChunker(self.id, get_text = lambda: text)

        yield from chunker.read()


    def to_dict(self) -> dict:
        return dict(
            id=str(self.id),
            type=self.type,
            title=self.title,
            file_path=self.file_path,
        )



from sqlalchemy import select
from sqlalchemy.orm import joinedload
from cognee.modules.data.models import Dataset
from cognee.infrastructure.databases.relational import get_relational_engine

async def ensure_dataset_exists(dataset_name: str) -> Dataset:
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        dataset = (await session.scalars(
            select(Dataset)\
                .options(joinedload(Dataset.data))\
                .filter(Dataset.name == dataset_name)
        )).first()

        if dataset is None:
            dataset = Dataset(
                name = dataset_name,
                data = []
            )

            session.add(dataset)

            await session.commit()

        return dataset



import asyncio
from cognee.shared.utils import render_graph
from cognee.infrastructure.databases.graph import get_graph_engine

if __name__ == "__main__":
    async def main():
        graph_client = await get_graph_engine()
        graph = graph_client.graph

        graph_url = await render_graph(graph)

        print(graph_url)

    asyncio.run(main())



from .sqlalchemy.SqlAlchemyAdapter import SQLAlchemyAdapter

def create_relational_engine(
    db_path: str,
    db_name: str,
    db_provider: str,
    db_host: str,
    db_port: str,
    db_user: str,
    db_password: str,
):
    return SQLAlchemyAdapter(
        db_name = db_name,
        db_path = db_path,
        db_type = db_provider,
        db_host = db_host,
        db_port = db_port,
        db_user = db_user,
        db_password = db_password
    )



import os
import uuid
from typing import Optional
from fastapi import Depends, Request
from fastapi_users import BaseUserManager, UUIDIDMixin
from fastapi_users.db import SQLAlchemyUserDatabase

from .get_user_db import get_user_db
from .models import User

class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    reset_password_token_secret = os.getenv("FASTAPI_USERS_RESET_PASSWORD_TOKEN_SECRET", "super_secret")
    verification_token_secret = os.getenv("FASTAPI_USERS_VERIFICATION_TOKEN_SECRET", "super_secret")

    async def on_after_register(self, user: User, request: Optional[Request] = None):
        print(f"User {user.id} has registered.")

    async def on_after_forgot_password(
        self, user: User, token: str, request: Optional[Request] = None
    ):
        print(f"User {user.id} has forgot their password. Reset token: {token}")

    async def on_after_request_verify(
        self, user: User, token: str, request: Optional[Request] = None
    ):
        print(f"Verification requested for user {user.id}. Verification token: {token}")

async def get_user_manager(user_db: SQLAlchemyUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)

from contextlib import asynccontextmanager
get_user_manager_context = asynccontextmanager(get_user_manager)



from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy.orm import relationship
from sqlalchemy import Column, DateTime, UUID
from cognee.infrastructure.databases.relational import Base
from .ACLResources import ACLResources

class Resource(Base):
    __tablename__ = "resources"

    id = Column(UUID(as_uuid = True), primary_key = True, default = uuid4)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))
    updated_at = Column(DateTime(timezone = True), onupdate = lambda: datetime.now(timezone.utc))

    resource_id = Column(UUID(as_uuid = True), nullable = False)

    acls = relationship("ACL", secondary = ACLResources.__tablename__, back_populates = "resources")



from cognee.infrastructure.databases.vector import get_vector_engine
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk


async def chunk_update_check(data_chunks: list[DocumentChunk], collection_name: str) -> list[DocumentChunk]:
    vector_engine = get_vector_engine()

    if not await vector_engine.has_collection(collection_name):
        # If collection doesn't exist, all data_chunks are new
        return data_chunks

    existing_chunks = await vector_engine.retrieve(
        collection_name,
        [str(chunk.chunk_id) for chunk in data_chunks],
    )

    existing_chunks_map = {chunk.id: chunk.payload for chunk in existing_chunks}

    affected_data_chunks = []

    for chunk in data_chunks:
        if chunk.chunk_id not in existing_chunks_map or \
        chunk.text != existing_chunks_map[chunk.chunk_id]["text"]:
            affected_data_chunks.append(chunk)

    return affected_data_chunks



import re

def chunk_by_word(data: str):
    sentence_endings = r"[.;!?…]"
    paragraph_endings = r"[\n\r]"
    last_processed_character = ""

    word = ""
    i = 0

    while i < len(data):
        character = data[i]

        if word == "" and (re.match(paragraph_endings, character) or character == " "):
            i = i + 1
            continue

        def is_real_paragraph_end():
            if re.match(sentence_endings, last_processed_character):
                return True

            j = i + 1
            next_character = data[j] if j < len(data) else None
            while next_character is not None and (re.match(paragraph_endings, next_character) or next_character == " "):
                j += 1
                next_character = data[j] if j < len(data) else None
            if next_character.isupper():
                return True

            return False

        if re.match(paragraph_endings, character):
            yield (word, "paragraph_end" if is_real_paragraph_end() else "word")
            word = ""
            i = i + 1
            continue

        if character == " ":
            yield [word, "word"]
            word = ""
            i = i + 1
            continue

        word += character
        last_processed_character = character

        if re.match(sentence_endings, character):
            # Check for ellipses.
            if i + 2 <= len(data) and data[i] == "." and data[i + 1] == "." and data[i + 2] == ".":
                word += ".."
                i = i + 2

            is_paragraph_end = i + 1 < len(data) and re.match(paragraph_endings, data[i + 1])
            yield (word, "paragraph_end" if is_paragraph_end else "sentence_end")
            word = ""

        i += 1

    if len(word) > 0:
        yield (word, "word")



from sqlalchemy.orm import joinedload
from sqlalchemy.future import select
from cognee.modules.users.models import User
from cognee.infrastructure.databases.relational import get_relational_engine

async def get_default_user():
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        query = select(User).options(joinedload(User.groups))\
            .where(User.email == "default_user@example.com")

        result = await session.execute(query)
        user = result.scalars().first()

        return user



from typing import List, Protocol, Optional
from abc import abstractmethod
from .models.DataPoint import DataPoint
from .models.PayloadSchema import PayloadSchema

class VectorDBInterface(Protocol):
    """ Collections """
    @abstractmethod
    async def has_collection(self, collection_name: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def create_collection(
        self,
        collection_name: str,
        payload_schema: Optional[PayloadSchema] = None,
    ): raise NotImplementedError

    """ Data points """
    @abstractmethod
    async def create_data_points(
        self,
        collection_name: str,
        data_points: List[DataPoint]
    ): raise NotImplementedError

    @abstractmethod
    async def retrieve(
        self,
        collection_name: str,
        data_point_ids: list[str]
    ): raise NotImplementedError

    """ Search """
    @abstractmethod
    async def search(
        self,
        collection_name: str,
        query_text: Optional[str],
        query_vector: Optional[List[float]],
        limit: int,
        with_vector: bool = False

    ): raise NotImplementedError

    @abstractmethod
    async def batch_search(
        self,
        collection_name: str,
        query_texts: List[str],
        limit: int,
        with_vectors: bool = False
    ): raise NotImplementedError

    @abstractmethod
    async def delete_data_points(
        self,
        collection_name: str,
        data_point_ids: list[str]
    ): raise NotImplementedError

    @abstractmethod
    async def prune(self): raise NotImplementedError




import logging



async def detect_language(data:str):
    """
    Detect the language of the given text and return its ISO 639-1 language code.
    If the detected language is Croatian ('hr'), it maps to Serbian ('sr').
    The text is trimmed to the first 100 characters for efficient processing.
    Parameters:
    text (str): The text for language detection.
    Returns:
    str: The ISO 639-1 language code of the detected language, or 'None' in case of an error.
    """

    # Trim the text to the first 100 characters
    from langdetect import detect, LangDetectException
    trimmed_text = data[:100]

    try:
        # Detect the language using langdetect
        detected_lang_iso639_1 = detect(trimmed_text)
        logging.info(f"Detected ISO 639-1 code: {detected_lang_iso639_1}")

        # Special case: map 'hr' (Croatian) to 'sr' (Serbian ISO 639-2)
        if detected_lang_iso639_1 == 'hr':
            yield 'sr'
        yield detected_lang_iso639_1

    except LangDetectException as e:
        logging.error(f"Language detection error: {e}")
    except Exception as e:
        logging.error(f"Unexpected error: {e}")

    yield None


from .get_auth_router import get_auth_router
from .get_register_router import get_register_router
from .get_reset_password_router import get_reset_password_router
from .get_users_router import get_users_router
from .get_verify_router import get_verify_router



from uuid import UUID, uuid5, NAMESPACE_OID
from cognee.modules.data.chunking.TextChunker import TextChunker
from .Document import Document

class TextDocument(Document):
    type: str = "text"
    title: str
    file_path: str

    def __init__(self, id: UUID, title: str, file_path: str):
        self.id = id or uuid5(NAMESPACE_OID, title)
        self.title = title
        self.file_path = file_path

    def read(self):
        def get_text():
            with open(self.file_path, mode = "r", encoding = "utf-8") as file:
                while True:
                    text = file.read(1024)

                    if len(text.strip()) == 0:
                        break

                    yield text


        chunker = TextChunker(self.id, get_text = get_text)

        yield from chunker.read()


    def to_dict(self) -> dict:
        return dict(
            id = str(self.id),
            type = self.type,
            title = self.title,
            file_path = self.file_path,
        )



from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import Column, ForeignKey, DateTime, UUID
from cognee.infrastructure.databases.relational import Base
from .ACLResources import ACLResources

class ACL(Base):
    __tablename__ = "acls"

    id = Column(UUID(as_uuid = True), primary_key = True, default = uuid4)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))
    updated_at = Column(DateTime(timezone = True), onupdate = lambda: datetime.now(timezone.utc))

    principal_id = Column(UUID(as_uuid = True), ForeignKey("principals.id"))
    permission_id = Column(UUID(as_uuid = True), ForeignKey("permissions.id"))

    principal = relationship("Principal")
    permission = relationship("Permission")
    resources: Mapped[list["Resource"]] = relationship(
        "Resource",
        secondary = ACLResources.__tablename__,
        back_populates = "acls",
    )



from __future__ import annotations
import re

from cognee.infrastructure.data.chunking.DefaultChunkEngine import DefaultChunkEngine
from cognee.shared.data_models import ChunkStrategy



class LangchainChunkEngine:
    def __init__(self, chunk_strategy=None, source_data=None, chunk_size=None, chunk_overlap=None):
        self.chunk_strategy = chunk_strategy
        self.source_data = source_data
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap



    def chunk_data(self,
        chunk_strategy = None,
        source_data = None,
        chunk_size = None,
        chunk_overlap = None,
    ):
        """
        Chunk data based on the specified strategy.

        Parameters:
        - chunk_strategy: The strategy to use for chunking.
        - source_data: The data to be chunked.
        - chunk_size: The size of each chunk.
        - chunk_overlap: The overlap between chunks.

        Returns:
        - The chunked data.
        """

        if chunk_strategy == ChunkStrategy.CODE:
            chunked_data, chunk_number = self.chunk_data_by_code(source_data,self.chunk_size, self.chunk_overlap)

        elif chunk_strategy == ChunkStrategy.LANGCHAIN_CHARACTER:
            chunked_data, chunk_number  = self.chunk_data_by_character(source_data,self.chunk_size, self.chunk_overlap)
        else:
            chunked_data, chunk_number  = "Invalid chunk strategy.", [0, "Invalid chunk strategy."]
        return chunked_data, chunk_number


    def chunk_data_by_code(self, data_chunks, chunk_size, chunk_overlap= 10, language=None):
        from langchain_text_splitters import (
            Language,
            RecursiveCharacterTextSplitter,
        )
        if language is None:
            language = Language.PYTHON
        python_splitter = RecursiveCharacterTextSplitter.from_language(
            language=language, chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )
        code_chunks = python_splitter.create_documents([data_chunks])

        only_content = [chunk.page_content for chunk in code_chunks]

        numbered_chunks = []
        for i, chunk in enumerate(code_chunks):
            numbered_chunk = [i + 1, chunk]
            numbered_chunks.append(numbered_chunk)

        return only_content, numbered_chunks

    def chunk_data_by_character(self, data_chunks, chunk_size=1500, chunk_overlap=10):
        from langchain_text_splitters import RecursiveCharacterTextSplitter
        splitter = RecursiveCharacterTextSplitter(chunk_size =chunk_size, chunk_overlap=chunk_overlap)
        data_chunks = splitter.create_documents([data_chunks])

        only_content = [chunk.page_content for chunk in data_chunks]

        numbered_chunks = []
        for i, chunk in enumerate(data_chunks):
            numbered_chunk = [i + 1, chunk]
            numbered_chunks.append(numbered_chunk)

        return only_content, numbered_chunks




from uuid import UUID
from cognee.infrastructure.databases.relational import get_relational_engine
from ..models.PipelineRun import PipelineRun

async def log_pipeline_status(run_id: UUID, status: str, run_info: dict):
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        session.add(PipelineRun(
            run_id = run_id,
            status = status,
            run_info = run_info,
        ))

        await session.commit()



from cognee.modules.data.models import Data
from cognee.modules.data.processing.document_types import Document, PdfDocument, AudioDocument, ImageDocument, TextDocument

def classify_documents(data_documents: list[Data]) -> list[Document]:
    documents = [
        PdfDocument(id = data_item.id, title=f"{data_item.name}.{data_item.extension}", file_path=data_item.raw_data_location) if data_item.extension == "pdf" else
        AudioDocument(id = data_item.id, title=f"{data_item.name}.{data_item.extension}", file_path=data_item.raw_data_location) if data_item.extension == "audio" else
        ImageDocument(id = data_item.id, title=f"{data_item.name}.{data_item.extension}", file_path=data_item.raw_data_location) if data_item.extension == "image" else
        TextDocument(id = data_item.id, title=f"{data_item.name}.{data_item.extension}", file_path=data_item.raw_data_location)
        for data_item in data_documents
    ]

    return documents



from typing import Union, Callable, Any, Coroutine, Generator, AsyncGenerator

class Task():
    executable: Union[
        Callable[..., Any],
        Callable[..., Coroutine[Any, Any, Any]],
        Generator[Any, Any, Any],
        AsyncGenerator[Any, Any],
    ]
    task_config: dict[str, Any] = {
        "batch_size": 1,
    }
    default_params: dict[str, Any] = {}

    def __init__(self, executable, *args, task_config = None, **kwargs):
        self.executable = executable
        self.default_params = {
            "args": args,
            "kwargs": kwargs
        }

        if task_config is not None:
            self.task_config = task_config

            if "batch_size" not in task_config:
                self.task_config["batch_size"] = 1

    def run(self, *args, **kwargs):
        combined_args = self.default_params["args"] + args
        combined_kwargs = { **self.default_params["kwargs"], **kwargs }

        return self.executable(*combined_args, **combined_kwargs)




import os
import logging
import pathlib
import cognee

logging.basicConfig(level=logging.DEBUG)

async def main():
    cognee.config.set_graph_database_provider("neo4j")
    data_directory_path = str(pathlib.Path(os.path.join(pathlib.Path(__file__).parent, ".data_storage/test_neo4j")).resolve())
    cognee.config.data_root_directory(data_directory_path)
    cognee_directory_path = str(pathlib.Path(os.path.join(pathlib.Path(__file__).parent, ".cognee_system/test_neo4j")).resolve())
    cognee.config.system_root_directory(cognee_directory_path)

    await cognee.prune.prune_data()
    await cognee.prune.prune_system(metadata = True)

    dataset_name = "cs_explanations"

    explanation_file_path = os.path.join(pathlib.Path(__file__).parent, "test_data/Natural_language_processing.txt")
    await cognee.add([explanation_file_path], dataset_name)

    text = """A quantum computer is a computer that takes advantage of quantum mechanical phenomena.
    At small scales, physical matter exhibits properties of both particles and waves, and quantum computing leverages this behavior, specifically quantum superposition and entanglement, using specialized hardware that supports the preparation and manipulation of quantum states.
    Classical physics cannot explain the operation of these quantum devices, and a scalable quantum computer could perform some calculations exponentially faster (with respect to input size scaling) than any modern "classical" computer. In particular, a large-scale quantum computer could break widely used encryption schemes and aid physicists in performing physical simulations; however, the current state of the technology is largely experimental and impractical, with several obstacles to useful applications. Moreover, scalable quantum computers do not hold promise for many practical tasks, and for many important tasks quantum speedups are proven impossible.
    The basic unit of information in quantum computing is the qubit, similar to the bit in traditional digital electronics. Unlike a classical bit, a qubit can exist in a superposition of its two "basis" states. When measuring a qubit, the result is a probabilistic output of a classical bit, therefore making quantum computers nondeterministic in general. If a quantum computer manipulates the qubit in a particular way, wave interference effects can amplify the desired measurement results. The design of quantum algorithms involves creating procedures that allow a quantum computer to perform calculations efficiently and quickly.
    Physically engineering high-quality qubits has proven challenging. If a physical qubit is not sufficiently isolated from its environment, it suffers from quantum decoherence, introducing noise into calculations. Paradoxically, perfectly isolating qubits is also undesirable because quantum computations typically need to initialize qubits, perform controlled qubit interactions, and measure the resulting quantum states. Each of those operations introduces errors and suffers from noise, and such inaccuracies accumulate.
    In principle, a non-quantum (classical) computer can solve the same computational problems as a quantum computer, given enough time. Quantum advantage comes in the form of time complexity rather than computability, and quantum complexity theory shows that some quantum algorithms for carefully selected tasks require exponentially fewer computational steps than the best known non-quantum algorithms. Such tasks can in theory be solved on a large-scale quantum computer whereas classical computers would not finish computations in any reasonable amount of time. However, quantum speedup is not universal or even typical across computational tasks, since basic tasks such as sorting are proven to not allow any asymptotic quantum speedup. Claims of quantum supremacy have drawn significant attention to the discipline, but are demonstrated on contrived tasks, while near-term practical use cases remain limited.
    """

    await cognee.add([text], dataset_name)

    await cognee.cognify([dataset_name])

    from cognee.infrastructure.databases.vector import get_vector_engine
    vector_engine = get_vector_engine()
    random_node = (await vector_engine.search("entities", "AI"))[0]
    random_node_name = random_node.payload["name"]

    search_results = await cognee.search("SIMILARITY", { "query": random_node_name })
    assert len(search_results) != 0, "The search results list is empty."
    print("\n\nExtracted sentences are:\n")
    for result in search_results:
        print(f"{result}\n")

    search_results = await cognee.search("TRAVERSE", { "query": random_node_name })
    assert len(search_results) != 0, "The search results list is empty."
    print("\n\nExtracted sentences are:\n")
    for result in search_results:
        print(f"{result}\n")

    search_results = await cognee.search("SUMMARY", { "query": random_node_name })
    assert len(search_results) != 0, "Query related summaries don't exist."
    print("\n\nQuery related summaries exist:\n")
    for result in search_results:
        print(f"{result}\n")

    search_results = await cognee.search("ADJACENT", { "query": random_node_name })
    assert len(search_results) != 0, "Large language model query found no neighbours."
    print("\n\Large language model query found neighbours.\n")
    for result in search_results:
        print(f"{result}\n")


if __name__ == "__main__":
    import asyncio
    asyncio.run(main())



from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy import Column, String, DateTime, UUID
from cognee.infrastructure.databases.relational import Base

class Principal(Base):
    __tablename__ = "principals"

    id = Column(UUID(as_uuid = True), primary_key = True, index = True, default = uuid4)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))
    updated_at = Column(DateTime(timezone = True), onupdate = lambda: datetime.now(timezone.utc))

    type = Column(String, nullable = False)

    __mapper_args__ = {
        "polymorphic_identity": "principal",
        "polymorphic_on": "type",
    }



from typing import Protocol, BinaryIO, Union


class IngestionData(Protocol):
    data: Union[str, BinaryIO] = None

    def get_data(self):
        raise NotImplementedError()

    def get_identifier(self):
        raise NotImplementedError()

    def get_metadata(self):
        raise NotImplementedError()



import uuid
from functools import lru_cache
from fastapi_users import FastAPIUsers
from .authentication.get_auth_backend import get_auth_backend

from .get_user_manager import get_user_manager
from .models.User import User

@lru_cache
def get_fastapi_users():
    auth_backend = get_auth_backend()

    fastapi_users = FastAPIUsers[User, uuid.UUID](get_user_manager, [auth_backend])

    return fastapi_users



from .config import get_vectordb_config
from .embeddings import get_embedding_engine
from .create_vector_engine import create_vector_engine

def get_vector_engine():
    return create_vector_engine(get_vectordb_config().to_dict(), get_embedding_engine())



from cognee.modules.data.deletion import prune_system, prune_data

class prune():
    @staticmethod
    async def prune_data():
        await prune_data()

    @staticmethod
    async def prune_system(graph = True, vector = True, metadata = False):
        await prune_system(graph, vector, metadata)

if __name__ == "__main__":
    import asyncio
    async def main():
        await prune.prune_data()
        await prune.prune_system()


    asyncio.run(main())



from fastapi_users.exceptions import UserAlreadyExists
from cognee.infrastructure.databases.relational import get_relational_engine
from ..get_user_manager import get_user_manager_context
from ..get_user_db import get_user_db_context
from ..models.User import UserCreate

async def create_user(
    email: str,
    password: str,
    is_superuser: bool = False,
    is_active: bool = True,
    is_verified: bool = False,
    auto_login: bool = False,
):
    try:
        relational_engine = get_relational_engine()

        async with relational_engine.get_async_session() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    user = await user_manager.create(
                        UserCreate(
                            email = email,
                            password = password,
                            is_superuser = is_superuser,
                            is_active = is_active,
                            is_verified = is_verified,
                        )
                    )

                    if auto_login:
                        await session.refresh(user)

                    return user
                    print(f"User created: {user.email}")
    except UserAlreadyExists as error:
        print(f"User {email} already exists")
        raise error



from pathlib import Path

ROOT_DIR = Path(__file__).resolve().parent

def get_absolute_path(path_from_root: str) -> str:
    absolute_path = ROOT_DIR / path_from_root
    return str(absolute_path.resolve())



""" LLM Interface """

from typing import Type, Protocol
from abc import abstractmethod
from pydantic import BaseModel
class LLMInterface(Protocol):
    """ LLM Interface """

    @abstractmethod
    async def acreate_structured_output(self,
                                        text_input: str,
                                        system_prompt: str,
                                        response_model: Type[BaseModel]) -> BaseModel:
        """To get structured output, import/call this function"""
        raise NotImplementedError

    @abstractmethod
    def show_prompt(self, text_input: str, system_prompt: str) -> str:
        """To get structured output, import/call this function"""
        raise NotImplementedError



from uuid import UUID
from sqlalchemy import select
from cognee.infrastructure.databases.relational import get_relational_engine
from ...models import ACL, Resource, Permission

async def get_document_ids_for_user(user_id: UUID) -> list[str]:
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        async with session.begin():
            document_ids = (await session.scalars(
                select(Resource.resource_id)
                .join(ACL.resources)
                .join(ACL.permission)
                .where(
                    ACL.principal_id == user_id,
                    Permission.name == "read",
                )
            )).all()

            return document_ids



from cognee.modules.data.processing.document_types import Document
from cognee.modules.users.permissions.methods import check_permission_on_documents

async def check_permissions_on_documents(documents: list[Document], user, permissions):
    document_ids = [document.id for document in documents]

    for permission in permissions:
        await check_permission_on_documents(
            user,
            permission,
            document_ids,
        )

    return documents



from cognee.infrastructure.databases.vector import get_vector_engine

async def search_summary(query: str) -> list:
    """
    Parameters:
    - query (str): The query string to filter summaries by.

    Returns:
    - list[str, UUID]: A list of objects providing information about the summaries related to query.
    """
    vector_engine = get_vector_engine()

    summaries_results = await vector_engine.search("chunk_summaries", query, limit = 5)

    summaries = [summary.payload for summary in summaries_results]

    return summaries



from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.shared.data_models import DefaultContentPrediction, SummarizedContent

class CognifyConfig(BaseSettings):
    classification_model: object = DefaultContentPrediction
    summarization_model: object = SummarizedContent

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "classification_model": self.classification_model,
            "summarization_model": self.summarization_model,
        }

@lru_cache
def get_cognify_config():
    return CognifyConfig()



# from fastapi import Depends
from fastapi_users.db import SQLAlchemyUserDatabase
# from cognee.infrastructure.databases.relational import get_relational_engine
from .models.User import User

async def get_user_db(session):
    yield SQLAlchemyUserDatabase(session, User)

from contextlib import asynccontextmanager
get_user_db_context = asynccontextmanager(get_user_db)



from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict

from cognee.infrastructure.data.chunking.DefaultChunkEngine import DefaultChunkEngine
from cognee.shared.data_models import ChunkStrategy, ChunkEngine


class ChunkConfig(BaseSettings):
    chunk_size: int = 1500
    chunk_overlap: int = 10
    chunk_strategy: object = ChunkStrategy.PARAGRAPH
    chunk_engine: object = ChunkEngine.DEFAULT_ENGINE


    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "chunk_size": self.chunk_size,
            "chunk_overlap": self.chunk_overlap,
            "chunk_strategy": self.chunk_strategy,
            "chunk_engine": self.chunk_engine,
        }


@lru_cache
def get_chunk_config():
    return ChunkConfig()



from cognee.modules.users.get_fastapi_users import get_fastapi_users
from cognee.modules.users.models.User import UserRead

def get_verify_router():
    return get_fastapi_users().get_verify_router(UserRead)



from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, Field

class RelationshipModel(BaseModel):
    type: str
    source: str
    target: str

class NodeModel(BaseModel):
    node_id: str
    name: str
    default_relationship: Optional[RelationshipModel] = None
    children: List[Union[Dict[str, Any], "NodeModel"]] = Field(default_factory=list)

NodeModel.update_forward_refs()


class OntologyNode(BaseModel):
    id: str = Field(..., description = "Unique identifier made from node name.")
    name: str
    description: str

class OntologyEdge(BaseModel):
    id: str
    source_id: str
    target_id: str
    relationship_type: str

class GraphOntology(BaseModel):
    nodes: list[OntologyNode]
    edges: list[OntologyEdge]



""" Neo4j Adapter for Graph Database"""
import json
import logging
import asyncio
from typing import Optional, Any, List, Dict
from contextlib import asynccontextmanager
from neo4j import AsyncSession
from neo4j import AsyncGraphDatabase
from neo4j.exceptions import Neo4jError
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

logger = logging.getLogger("Neo4jAdapter")

class Neo4jAdapter(GraphDBInterface):
    def __init__(
        self,
        graph_database_url: str,
        graph_database_username: str,
        graph_database_password: str,
        driver: Optional[Any] = None,
    ):
        self.driver = driver or AsyncGraphDatabase.driver(
            graph_database_url,
            auth = (graph_database_username, graph_database_password),
            max_connection_lifetime = 120
        )

    async def close(self) -> None:
        await self.driver.close()

    @asynccontextmanager
    async def get_session(self) -> AsyncSession:
        async with self.driver.session() as session:
            yield session

    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        try:
            async with self.get_session() as session:
                result = await session.run(query, parameters=params)
                data = await result.data()
                await self.close()
                return data
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info = True)
            raise error

    async def graph(self):
        return await self.get_session()

    async def has_node(self, node_id: str) -> bool:
        results = self.query(
            """
                MATCH (n)
                WHERE n.id = $node_id
                RETURN COUNT(n) > 0 AS node_exists
            """,
            {"node_id": node_id}
        )
        return results[0]["node_exists"] if len(results) > 0 else False

    async def add_node(self, node_id: str, node_properties: Dict[str, Any] = None):
        node_id = node_id.replace(":", "_")

        serialized_properties = self.serialize_properties(node_properties)

        if "name" not in serialized_properties:
            serialized_properties["name"] = node_id

        query = f"""MERGE (node:`{node_id}` {{id: $node_id}})
                ON CREATE SET node += $properties
                RETURN ID(node) AS internal_id, node.id AS nodeId"""

        params = {
            "node_id": node_id,
            "properties": serialized_properties,
        }

        return await self.query(query, params)

    async def add_nodes(self, nodes: list[tuple[str, dict[str, Any]]]) -> None:
        query = """
        UNWIND $nodes AS node
        MERGE (n {id: node.node_id})
        ON CREATE SET n += node.properties
        WITH n, node.node_id AS label
        CALL apoc.create.addLabels(n, [label]) YIELD node AS labeledNode
        RETURN ID(labeledNode) AS internal_id, labeledNode.id AS nodeId
        """

        nodes = [{
            "node_id": node_id,
            "properties": self.serialize_properties(node_properties),
        } for (node_id, node_properties) in nodes]

        results = await self.query(query, dict(nodes = nodes))
        return results

    async def extract_node_description(self, node_id: str):
        query = """MATCH (n)-[r]->(m)
                    WHERE n.id = $node_id
                    AND NOT m.id CONTAINS 'DefaultGraphModel'
                    RETURN m
                    """

        result = await self.query(query, dict(node_id = node_id))

        descriptions = []

        for node in result:
            # Assuming 'm' is a consistent key in your data structure
            attributes = node.get("m", {})

            # Ensure all required attributes are present
            if all(key in attributes for key in ["id", "layer_id", "description"]):
                descriptions.append({
                    "id": attributes["id"],
                    "layer_id": attributes["layer_id"],
                    "description": attributes["description"],
                })

        return descriptions

    async def get_layer_nodes(self):
        query = """MATCH (node) WHERE node.layer_id IS NOT NULL
        RETURN node"""

        return [result["node"] for result in (await self.query(query))]

    async def extract_node(self, node_id: str):
        results = await self.extract_nodes([node_id])

        return results[0] if len(results) > 0 else None

    async def extract_nodes(self, node_ids: List[str]):
        query = """
        UNWIND $node_ids AS id
        MATCH (node {id: id})
        RETURN node"""

        params = {
            "node_ids": node_ids
        }

        results = await self.query(query, params)

        return [result["node"] for result in results]

    async def delete_node(self, node_id: str):
        node_id = id.replace(":", "_")

        query = f"MATCH (node:`{node_id}` {{id: $node_id}}) DETACH DELETE n"
        params = { "node_id": node_id }

        return await self.query(query, params)

    async def delete_nodes(self, node_ids: list[str]) -> None:
        query = """
        UNWIND $node_ids AS id
        MATCH (node {id: id})
        DETACH DELETE node"""

        params = {
            "node_ids": node_ids
        }

        return await self.query(query, params)

    async def has_edge(self, from_node: str, to_node: str, edge_label: str) -> bool:
        query = f"""
            MATCH (from_node:`{from_node}`)-[relationship:`{edge_label}`]->(to_node:`{to_node}`)
            RETURN COUNT(relationship) > 0 AS edge_exists
        """

        edge_exists = await self.query(query)
        return edge_exists

    async def has_edges(self, edges):
        query = """
            UNWIND $edges AS edge
            MATCH (a)-[r]->(b)
            WHERE id(a) = edge.from_node AND id(b) = edge.to_node AND type(r) = edge.relationship_name
            RETURN edge.from_node AS from_node, edge.to_node AS to_node, edge.relationship_name AS relationship_name, count(r) > 0 AS edge_exists
        """

        try:
            params = {
                "edges": [{
                    "from_node": edge[0],
                    "to_node": edge[1],
                    "relationship_name": edge[2],
                } for edge in edges],
            }

            results = await self.query(query, params)
            return [result["edge_exists"] for result in results]
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info = True)
            raise error


    async def add_edge(self, from_node: str, to_node: str, relationship_name: str, edge_properties: Optional[Dict[str, Any]] = {}):
        serialized_properties = self.serialize_properties(edge_properties)
        from_node = from_node.replace(":", "_")
        to_node = to_node.replace(":", "_")

        query = f"""MATCH (from_node:`{from_node}`
         {{id: $from_node}}), 
         (to_node:`{to_node}` {{id: $to_node}})
         MERGE (from_node)-[r:`{relationship_name}`]->(to_node)
         SET r += $properties
         RETURN r"""

        params = {
            "from_node": from_node,
            "to_node": to_node,
            "properties": serialized_properties
        }

        return await self.query(query, params)


    async def add_edges(self, edges: list[tuple[str, str, str, dict[str, Any]]]) -> None:
        query = """
        UNWIND $edges AS edge
        MATCH (from_node {id: edge.from_node})
        MATCH (to_node {id: edge.to_node})
        CALL apoc.create.relationship(from_node, edge.relationship_name, edge.properties, to_node) YIELD rel
        RETURN rel
        """

        edges = [{
          "from_node": edge[0],
          "to_node": edge[1],
          "relationship_name": edge[2],
          "properties": {
              **(edge[3] if edge[3] else {}),
              "source_node_id": edge[0],
              "target_node_id": edge[1],
          },
        } for edge in edges]

        try:
            results = await self.query(query, dict(edges = edges))
            return results
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info = True)
            raise error

    async def get_edges(self, node_id: str):
        query = """
        MATCH (n {id: $node_id})-[r]-(m)
        RETURN n, r, m
        """

        results = await self.query(query, dict(node_id = node_id))

        return [(result["n"]["id"], result["m"]["id"], {"relationship_name": result["r"][1]}) for result in results]

    async def get_disconnected_nodes(self) -> list[str]:
        # return await self.query(
        #     "MATCH (node) WHERE NOT (node)<-[:*]-() RETURN node.id as id",
        # )
        query = """
        // Step 1: Collect all nodes
        MATCH (n)
        WITH COLLECT(n) AS nodes

        // Step 2: Find all connected components
        WITH nodes
        CALL {
          WITH nodes
          UNWIND nodes AS startNode
          MATCH path = (startNode)-[*]-(connectedNode)
          WITH startNode, COLLECT(DISTINCT connectedNode) AS component
          RETURN component
        }

        // Step 3: Aggregate components
        WITH COLLECT(component) AS components

        // Step 4: Identify the largest connected component
        UNWIND components AS component
        WITH component
        ORDER BY SIZE(component) DESC
        LIMIT 1
        WITH component AS largestComponent

        // Step 5: Find nodes not in the largest connected component
        MATCH (n)
        WHERE NOT n IN largestComponent
        RETURN COLLECT(ID(n)) AS ids
        """

        results = await self.query(query)
        return results[0]["ids"] if len(results) > 0 else []


    async def filter_nodes(self, search_criteria):
        query = f"""MATCH (node)
                WHERE node.id CONTAINS '{search_criteria}'
                RETURN node"""

        return await self.query(query)


    async def get_predecessor_ids(self, node_id: str, edge_label: str = None) -> list[str]:
        if edge_label is not None:
            query = """
            MATCH (node)<-[r]-(predecessor)
            WHERE node.id = $node_id AND type(r) = $edge_label
            RETURN predecessor.id AS predecessor_id
            """

            results = await self.query(
                query,
                dict(
                    node_id = node_id,
                    edge_label = edge_label,
                )
            )

            return [result["predecessor_id"] for result in results]
        else:
            query = """
            MATCH (node)<-[r]-(predecessor)
            WHERE node.id = $node_id
            RETURN predecessor.id AS predecessor_id
            """

            results = await self.query(
                query,
                dict(
                    node_id = node_id,
                )
            )

            return [result["predecessor_id"] for result in results]

    async def get_successor_ids(self, node_id: str, edge_label: str = None) -> list[str]:
        if edge_label is not None:
            query = """
            MATCH (node)-[r]->(successor)
            WHERE node.id = $node_id AND type(r) = $edge_label
            RETURN successor.id AS successor_id
            """

            results = await self.query(
                query,
                dict(
                    node_id = node_id,
                    edge_label = edge_label,
                ),
            )

            return [result["successor_id"] for result in results]
        else:
            query = """
            MATCH (node)-[r]->(successor)
            WHERE node.id = $node_id
            RETURN successor.id AS successor_id
            """

            results = await self.query(
                query,
                dict(
                    node_id = node_id,
                )
            )

            return [result["successor_id"] for result in results]

    async def get_neighbours(self, node_id: str) -> list[str]:
        predecessor_ids, successor_ids = await asyncio.gather(self.get_predecessor_ids(node_id), self.get_successor_ids(node_id))

        return [*predecessor_ids, *successor_ids]

    async def remove_connection_to_predecessors_of(self, node_ids: list[str], edge_label: str) -> None:
        query = f"""
        UNWIND $node_ids AS id
        MATCH (node:`{id}`)-[r:{edge_label}]->(predecessor)
        DELETE r;
        """

        params = { "node_ids": node_ids }

        return await self.query(query, params)

    async def remove_connection_to_successors_of(self, node_ids: list[str], edge_label: str) -> None:
        query = f"""
        UNWIND $node_ids AS id
        MATCH (node:`{id}`)<-[r:{edge_label}]-(successor)
        DELETE r;
        """

        params = { "node_ids": node_ids }

        return await self.query(query, params)


    async def delete_graph(self):
        query = """MATCH (node)
                DETACH DELETE node;"""

        return await self.query(query)

    def serialize_properties(self, properties = dict()):
        return {
            property_key: json.dumps(property_value)
            if isinstance(property_value, (dict, list))
            else property_value for property_key, property_value in properties.items()
        }

    async def get_graph_data(self):
        query = "MATCH (n) RETURN ID(n) AS id, labels(n) AS labels, properties(n) AS properties"
        result = await self.query(query)
        nodes = [(
            record["properties"]["id"],
            record["properties"],
        ) for record in result]

        query = """
        MATCH (n)-[r]->(m)
        RETURN ID(n) AS source, ID(m) AS target, TYPE(r) AS type, properties(r) AS properties
        """
        result = await self.query(query)
        edges = [(
            record["properties"]["source_node_id"],
            record["properties"]["target_node_id"],
            record["type"],
            record["properties"],
        ) for record in result]

        return (nodes, edges)



from typing import Dict


from cognee.shared.data_models import ChunkEngine


class ChunkingConfig(Dict):
    vector_db_url: str
    vector_db_key: str
    vector_db_provider: str

def create_chunking_engine(config: ChunkingConfig):
    if config["chunk_engine"] == ChunkEngine.LANGCHAIN_ENGINE:
        from cognee.infrastructure.data.chunking.LangchainChunkingEngine import LangchainChunkEngine
        return LangchainChunkEngine(
            chunk_size=config["chunk_size"],
            chunk_overlap=config["chunk_overlap"],
            chunk_strategy=config["chunk_strategy"],
        )
    elif config["chunk_engine"] == ChunkEngine.DEFAULT_ENGINE:
        from cognee.infrastructure.data.chunking.DefaultChunkEngine import DefaultChunkEngine

        return DefaultChunkEngine(
            chunk_size=config["chunk_size"],
            chunk_overlap=config["chunk_overlap"],
            chunk_strategy=config["chunk_strategy"],
        )
    elif config["chunk_engine"] == ChunkEngine.HAYSTACK_ENGINE:
        from cognee.infrastructure.data.chunking.HaystackChunkEngine import HaystackChunkEngine

        return HaystackChunkEngine(
            chunk_size=config["chunk_size"],
            chunk_overlap=config["chunk_overlap"],
            chunk_strategy=config["chunk_strategy"],
        )



import  logging
from uuid import UUID
from sqlalchemy import select
from sqlalchemy.orm import joinedload
from cognee.infrastructure.databases.relational import get_relational_engine

from ...models.User import User
from ...models.ACL import ACL

logger = logging.getLogger(__name__)

class PermissionDeniedException(Exception):
    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


async def check_permission_on_documents(user: User, permission_type: str, document_ids: list[UUID]):
    try:
        user_group_ids = [group.id for group in user.groups]

        db_engine = get_relational_engine()

        async with db_engine.get_async_session() as session:
            result = await session.execute(
                select(ACL)
                .join(ACL.permission)
                .options(joinedload(ACL.resources))
                .where(ACL.principal_id.in_([user.id, *user_group_ids]))
                .where(ACL.permission.has(name = permission_type))
            )
            acls = result.unique().scalars().all()
            resource_ids = [resource.resource_id for acl in acls for resource in acl.resources]
            has_permissions = all(document_id in resource_ids for document_id in document_ids)

            if not has_permissions:
                raise PermissionDeniedException(f"User {user.username} does not have {permission_type} permission on documents")
    except Exception as error:
        logger.error("Error checking permissions on documents: %s", str(error))
        raise



from uuid import uuid4
from datetime import datetime, timezone
from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import Column, String, DateTime, UUID, Text
from cognee.infrastructure.databases.relational import Base
from .PipelineTask import PipelineTask

class Task(Base):
    __tablename__ = "tasks"

    id = Column(UUID(as_uuid = True), primary_key = True, default = uuid4)

    name = Column(String)
    description = Column(Text, nullable = True)

    executable = Column(Text)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))
    updated_at = Column(DateTime(timezone = True), onupdate = lambda: datetime.now(timezone.utc))

    datasets: Mapped[list["Pipeline"]] = relationship(
        secondary = PipelineTask.__tablename__,
        back_populates = "task"
    )



from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import JSONResponse
from sqlalchemy.orm import Session
from cognee.modules.users import get_user_db
from cognee.modules.users.models import User, Group, Permission

def get_permissions_router() -> APIRouter:
    permissions_router = APIRouter()

    @permissions_router.post("/groups/{group_id}/permissions")
    async def give_permission_to_group(group_id: int, permission: str, db: Session = Depends(get_user_db)):
        group = db.query(Group).filter(Group.id == group_id).first()

        if not group:
            raise HTTPException(status_code = 404, detail = "Group not found")

        permission = db.query(Permission).filter(Permission.name == permission).first()

        if not permission:
            permission = Permission(name = permission)
            db.add(permission)

        group.permissions.append(permission)

        db.commit()

        return JSONResponse(status_code = 200, content = {"message": "Permission assigned to group"})

    @permissions_router.post("/users/{user_id}/groups")
    async def add_user_to_group(user_id: int, group_id: int, db: Session = Depends(get_user_db)):
        user = db.query(User).filter(User.id == user_id).first()
        group = db.query(Group).filter(Group.id == group_id).first()

        if not user or not group:
            raise HTTPException(status_code = 404, detail = "User or group not found")

        user.groups.append(group)

        db.commit()

        return JSONResponse(status_code = 200, content = {"message": "User added to group"})

    return permissions_router



""" This module contains the OntologyEngine class which is responsible for adding graph ontology from a JSON or CSV file. """

import csv
import json
import logging
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Union, Type

import aiofiles
import pandas as pd
from pydantic import BaseModel

from cognee.infrastructure.llm.prompts import read_query_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client
from cognee.infrastructure.data.chunking.config import get_chunk_config
from cognee.infrastructure.data.chunking.get_chunking_engine import get_chunk_engine
from cognee.infrastructure.databases.graph.get_graph_engine import get_graph_engine
from cognee.infrastructure.files.utils.extract_text_from_file import extract_text_from_file
from cognee.infrastructure.files.utils.guess_file_type import guess_file_type, FileTypeException
from cognee.modules.data.extraction.knowledge_graph.add_model_class_to_graph import add_model_class_to_graph
from cognee.tasks.infer_data_ontology.models.models import NodeModel, GraphOntology
from cognee.shared.data_models import KnowledgeGraph
from cognee.modules.graph.utils import generate_node_id, generate_node_name

logger = logging.getLogger("task:infer_data_ontology")


async def extract_ontology(content: str, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    system_prompt = read_query_prompt("extract_ontology.txt")

    ontology = await llm_client.acreate_structured_output(content, system_prompt, response_model)

    return ontology


class OntologyEngine:
    async def flatten_model(self, model: NodeModel, parent_id: Optional[str] = None) -> Dict[str, Any]:
        """Flatten the model to a dictionary."""
        result = model.dict()
        result["parent_id"] = parent_id
        if model.default_relationship:
            result.update({
                "relationship_type": model.default_relationship.type,
                "relationship_source": model.default_relationship.source,
                "relationship_target": model.default_relationship.target
            })
        return result

    async def recursive_flatten(self, items: Union[List[Dict[str, Any]], Dict[str, Any]], parent_id: Optional[str] = None) -> List[Dict[str, Any]]:
        """Recursively flatten the items.  """
        flat_list = []

        if isinstance(items, list):
            for item in items:
                flat_list.extend(await self.recursive_flatten(item, parent_id))
        elif isinstance(items, dict):
            model = NodeModel.parse_obj(items)
            flat_list.append(await self.flatten_model(model, parent_id))
            for child in model.children:
                flat_list.extend(await self.recursive_flatten(child, model.node_id))
        return flat_list

    async def load_data(self, file_path: str) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Load data from a JSON or CSV file."""
        try:
            if file_path.endswith(".json"):
                async with aiofiles.open(file_path, mode="r") as f:
                    data = await f.read()
                    return json.loads(data)
            elif file_path.endswith(".csv"):
                async with aiofiles.open(file_path, mode="r") as f:
                    content = await f.read()
                    reader = csv.DictReader(content.splitlines())
                    return list(reader)
            else:
                raise ValueError("Unsupported file format")
        except Exception as e:
            raise RuntimeError(f"Failed to load data from {file_path}: {e}")

    async def add_graph_ontology(self, file_path: str = None, documents: list = None):
        """Add graph ontology from a JSON or CSV file or infer from documents content."""
        if file_path is None:
            initial_chunks_and_ids = []

            chunk_config = get_chunk_config()
            chunk_engine = get_chunk_engine()
            chunk_strategy = chunk_config.chunk_strategy

            for base_file in documents:
                with open(base_file.file_path, "rb") as file:
                    try:
                        file_type = guess_file_type(file)
                        text = extract_text_from_file(file, file_type)

                        subchunks, chunks_with_ids = chunk_engine.chunk_data(
                            chunk_strategy,
                            text,
                            chunk_config.chunk_size,
                            chunk_config.chunk_overlap,
                        )

                        if chunks_with_ids[0][0] == 1:
                            initial_chunks_and_ids.append({base_file.id: chunks_with_ids})

                    except FileTypeException:
                        logger.warning("File (%s) has an unknown file type. We are skipping it.", file["id"])


            ontology = await extract_ontology(str(initial_chunks_and_ids), GraphOntology)
            graph_client = await get_graph_engine()

            await graph_client.add_nodes([(node.id, dict(
                uuid = generate_node_id(node.id),
                name = generate_node_name(node.name),
                type = generate_node_id(node.id),
                description = node.description,
                created_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                updated_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
            )) for node in ontology.nodes])

            await graph_client.add_edges((
                generate_node_id(edge.source_id),
                generate_node_id(edge.target_id),
                edge.relationship_type,
                dict(
                    source_node_id = generate_node_id(edge.source_id),
                    target_node_id = generate_node_id(edge.target_id),
                    relationship_name = edge.relationship_type,
                    created_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                    updated_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                ),
            ) for edge in ontology.edges)

        else:
            dataset_level_information = documents[0][1]

            # Extract the list of valid IDs from the explanations
            valid_ids = {item["id"] for item in dataset_level_information}
            try:
                data = await self.load_data(file_path)
                flt_ontology = await self.recursive_flatten(data)
                df = pd.DataFrame(flt_ontology)
                graph_client = await get_graph_engine()

                for _, row in df.iterrows():
                    node_data = row.to_dict()
                    node_id = node_data.pop("node_id", None)
                    if node_id in valid_ids:
                        await graph_client.add_node(node_id, node_data)
                    if node_id not in valid_ids:
                        raise ValueError(f"Node ID {node_id} not found in the dataset")
                    if pd.notna(row.get("relationship_source")) and pd.notna(row.get("relationship_target")):
                        await graph_client.add_edge(
                            row["relationship_source"],
                            row["relationship_target"],
                            relationship_name=row["relationship_type"],
                            edge_properties = {
                                "source_node_id": row["relationship_source"],
                                "target_node_id": row["relationship_target"],
                                "relationship_name": row["relationship_type"],
                                "created_at": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                                "updated_at": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                            },
                        )

                return
            except Exception as e:
                raise RuntimeError(f"Failed to add graph ontology from {file_path}: {e}") from e


async def infer_data_ontology(documents, ontology_model = KnowledgeGraph, root_node_id = None):
    if ontology_model == KnowledgeGraph:
        ontology_engine = OntologyEngine()
        root_node_id = await ontology_engine.add_graph_ontology(documents = documents)
    else:
        graph_engine = get_graph_engine()
        await add_model_class_to_graph(ontology_model, graph_engine)

    yield (documents, root_node_id)



from cognee.infrastructure.databases.vector import get_vectordb_config
from cognee.infrastructure.llm import get_llm_config

def get_settings():
    llm_config = get_llm_config()

    vector_dbs = [{
        "value": "weaviate",
        "label": "Weaviate",
    }, {
        "value": "qdrant",
        "label": "Qdrant",
    }, {
        "value": "lancedb",
        "label": "LanceDB",
    }]

    vector_config = get_vectordb_config()

    llm_providers = [{
        "value": "openai",
        "label": "OpenAI",
    }, {
        "value": "ollama",
        "label": "Ollama",
    }, {
        "value": "anthropic",
        "label": "Anthropic",
    }]

    llm_config = get_llm_config()

    return dict(
        llm = {
            "provider": {
                "label": llm_config.llm_provider,
                "value": llm_config.llm_provider,
            } if llm_config.llm_provider else llm_providers[0],
            "model": {
                "value": llm_config.llm_model,
                "label": llm_config.llm_model,
            } if llm_config.llm_model else None,
            "apiKey": (llm_config.llm_api_key[:-10] + "**********") if llm_config.llm_api_key else None,
            "providers": llm_providers,
            "models": {
                "openai": [{
                    "value": "gpt-4o",
                    "label": "gpt-4o",
                }, {
                    "value": "gpt-4-turbo",
                    "label": "gpt-4-turbo",
                }, {
                    "value": "gpt-3.5-turbo",
                    "label": "gpt-3.5-turbo",
                }],
                "ollama": [{
                    "value": "llama3",
                    "label": "llama3",
                }, {
                    "value": "mistral",
                    "label": "mistral",
                }],
                "anthropic": [{
                    "value": "Claude 3 Opus",
                    "label": "Claude 3 Opus",
                }, {
                    "value": "Claude 3 Sonnet",
                    "label": "Claude 3 Sonnet",
                }, {
                    "value": "Claude 3 Haiku",
                    "label": "Claude 3 Haiku",
                }]
            },
        },
        vectorDB = {
            "provider": {
                "label": vector_config.vector_engine_provider,
                "value": vector_config.vector_engine_provider.lower(),
            },
            "url": vector_config.vector_db_url,
            "apiKey": vector_config.vector_db_key,
            "options": vector_dbs,
        },
    )



from datetime import datetime, timezone
from sqlalchemy import Column, ForeignKey, UUID, DateTime
from cognee.infrastructure.databases.relational import Base

class ACLResources(Base):
    __tablename__ = "acl_resources"

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))

    acl_id = Column(UUID(as_uuid = True), ForeignKey("acls.id"), primary_key = True)
    resource_id = Column(UUID(as_uuid = True), ForeignKey("resources.id"), primary_key = True)




from cognee.infrastructure.databases.graph import get_graph_engine, get_graph_config

async def search_cypher(query: str):
    """
    Use a Cypher query to search the graph and return the results.
    """
    graph_config = get_graph_config()

    if graph_config.graph_database_provider == "neo4j":
        graph_engine = await get_graph_engine()
        result = await graph_engine.graph().run(query)
        return result
    else:
        raise ValueError("Unsupported search type for the used graph engine.")



from uuid import UUID, uuid4
from typing import Optional
from pydantic import BaseModel
from .models.Task import Task

class PipelineConfig(BaseModel):
    batch_count: int = 10
    description: Optional[str]

class Pipeline():
    id: UUID = uuid4()
    name: str
    description: str
    tasks: list[Task] = []

    def __init__(self, name: str, pipeline_config: PipelineConfig):
        self.name = name
        self.description = pipeline_config.description



from cognee.infrastructure.databases.vector import get_vector_engine
from .chunk_types import DocumentChunk

async def has_new_chunks(data_chunks: list[DocumentChunk], collection_name: str) -> list[DocumentChunk]:
    vector_engine = get_vector_engine()

    if not await vector_engine.has_collection(collection_name):
        # There is no collection created,
        # so no existing chunks, all chunks are new.
        return True

    existing_chunks = await vector_engine.retrieve(
        collection_name,
        [str(chunk.chunk_id) for chunk in data_chunks],
    )

    if len(existing_chunks) == 0:
        # If we don't find any existing chunk,
        # all chunks are new.
        return True

    existing_chunks_map = {chunk.id: chunk.payload for chunk in existing_chunks}

    new_data_chunks = [
        chunk for chunk in data_chunks \
            if chunk.chunk_id not in existing_chunks_map \
            or chunk.text != existing_chunks_map[chunk.chunk_id]["text"]
    ]

    return len(new_data_chunks) > 0



from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.modules.data.processing.document_types.Document import Document


async def source_documents_to_chunks(documents: list[Document], parent_node_id: str = None):
    graph_engine = await get_graph_engine()

    if parent_node_id is None:
        documents, parent_node_id = documents


    nodes = []
    edges = []

    if parent_node_id and await graph_engine.extract_node(parent_node_id) is None:
        nodes.append((parent_node_id, {}))

    document_nodes = await graph_engine.extract_nodes([str(document.id) for document in documents])

    for (document_index, document) in enumerate(documents):
        document_node = document_nodes[document_index] if document_index in document_nodes else None

        if document_node is None:
            nodes.append((str(document.id), document.to_dict()))

            if parent_node_id:
                edges.append((
                  parent_node_id,
                  str(document.id),
                  "has_document",
                  dict(
                      relationship_name = "has_document",
                      source_node_id = parent_node_id,
                      target_node_id = str(document.id),
                  ),
                ))

    if len(nodes) > 0:
        await graph_engine.add_nodes(nodes)
        await graph_engine.add_edges(edges)

    for document in documents:
        for document_chunk in document.read():
            yield document_chunk



from cognee.tasks.chunking import chunk_by_paragraph

if __name__ == "__main__":
    def test_chunking_on_whole_text():
        test_text = """This is example text. It contains multiple sentences.
        This is a second paragraph. First two paragraphs are whole.
        Third paragraph is a bit longer and is finished with a dot."""

        chunks = []

        for chunk_data in chunk_by_paragraph(test_text, 12, batch_paragraphs = False):
            chunks.append(chunk_data)

        assert len(chunks) == 3

        assert chunks[0]["text"] == "This is example text. It contains multiple sentences."
        assert chunks[0]["word_count"] == 8
        assert chunks[0]["cut_type"] == "paragraph_end"

        assert chunks[1]["text"] == "This is a second paragraph. First two paragraphs are whole."
        assert chunks[1]["word_count"] == 10
        assert chunks[1]["cut_type"] == "paragraph_end"

        assert chunks[2]["text"] == "Third paragraph is a bit longer and is finished with a dot."
        assert chunks[2]["word_count"] == 12
        assert chunks[2]["cut_type"] == "sentence_end"

    def test_chunking_on_cut_text():
        test_text = """This is example text. It contains multiple sentences.
        This is a second paragraph. First two paragraphs are whole.
        Third paragraph is cut and is missing the dot at the end"""

        chunks = []

        for chunk_data in chunk_by_paragraph(test_text, 12, batch_paragraphs = False):
            chunks.append(chunk_data)

        assert len(chunks) == 3

        assert chunks[0]["text"] == "This is example text. It contains multiple sentences."
        assert chunks[0]["word_count"] == 8
        assert chunks[0]["cut_type"] == "paragraph_end"

        assert chunks[1]["text"] == "This is a second paragraph. First two paragraphs are whole."
        assert chunks[1]["word_count"] == 10
        assert chunks[1]["cut_type"] == "paragraph_end"

        assert chunks[2]["text"] == "Third paragraph is cut and is missing the dot at the end"
        assert chunks[2]["word_count"] == 12
        assert chunks[2]["cut_type"] == "sentence_cut"

    test_chunking_on_whole_text()
    test_chunking_on_cut_text()



from uuid import UUID as uuid_UUID
from sqlalchemy import ForeignKey, UUID, Column
from sqlalchemy.orm import relationship, Mapped
from fastapi_users.db import SQLAlchemyBaseUserTableUUID
from .Principal import Principal
from .UserGroup import UserGroup

class User(SQLAlchemyBaseUserTableUUID, Principal):
    __tablename__ = "users"

    id = Column(UUID(as_uuid = True), ForeignKey("principals.id"), primary_key = True)

    groups: Mapped[list["Group"]] = relationship(
        secondary = UserGroup.__tablename__,
        back_populates = "users",
    )

    __mapper_args__ = {
        "polymorphic_identity": "user",
    }


# Keep these schemas in sync with User model
from fastapi_users import schemas

class UserRead(schemas.BaseUser[uuid_UUID]):
    groups: list[uuid_UUID]  # Add groups attribute

class UserCreate(schemas.BaseUserCreate):
    pass

class UserUpdate(schemas.BaseUserUpdate):
    pass



""" This module contains the configuration for the graph database. """

import os
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.shared.data_models import KnowledgeGraph
from cognee.root_dir import get_absolute_path


class GraphConfig(BaseSettings):
    graph_filename: str = "cognee_graph.pkl"
    graph_database_provider: str = "NETWORKX"
    graph_database_url: str = ""
    graph_database_username: str = ""
    graph_database_password: str = ""
    graph_database_port: int = 123
    graph_file_path: str = os.path.join(
        os.path.join(get_absolute_path(".cognee_system"), "databases"),
        graph_filename
    )
    graph_model: object = KnowledgeGraph
    graph_topology: object = KnowledgeGraph
    model_config = SettingsConfigDict(env_file=".env", extra="allow")

    def to_dict(self) -> dict:
        return {
            "graph_filename": self.graph_filename,
            "graph_database_provider": self.graph_database_provider,
            "graph_file_path": self.graph_file_path,
            "graph_database_url": self.graph_database_url,
            "graph_database_username": self.graph_database_username,
            "graph_database_password": self.graph_database_password,
            "graph_database_port": self.graph_database_port,
        }


@lru_cache
def get_graph_config():
    return GraphConfig()



import os
from functools import lru_cache
from fastapi_users import models
from fastapi_users.authentication import (
    AuthenticationBackend,
    BearerTransport,
    JWTStrategy,
)

@lru_cache
def get_auth_backend():
    bearer_transport = BearerTransport(tokenUrl = "auth/jwt/login")

    def get_jwt_strategy() -> JWTStrategy[models.UP, models.ID]:
        secret = os.getenv("FASTAPI_USERS_JWT_SECRET", "super_secret")
        return JWTStrategy(secret, lifetime_seconds = 3600)

    auth_backend = AuthenticationBackend(
        name = "jwt",
        transport = bearer_transport,
        get_strategy = get_jwt_strategy,
    )

    return auth_backend




import logging

from cognee.base_config import get_base_config

BaseConfig = get_base_config()

async def translate_text(data, source_language:str='sr', target_language:str='en', region_name='eu-west-1'):
    """
    Translate text from source language to target language using AWS Translate.
    Parameters:
    data (str): The text to be translated.
    source_language (str): The source language code (e.g., 'sr' for Serbian). ISO 639-2 Code https://www.loc.gov/standards/iso639-2/php/code_list.php
    target_language (str): The target language code (e.g., 'en' for English). ISO 639-2 Code https://www.loc.gov/standards/iso639-2/php/code_list.php
    region_name (str): AWS region name.
    Returns:
    str: Translated text or an error message.
    """
    import boto3
    from botocore.exceptions import BotoCoreError, ClientError

    if not data:
        yield "No text provided for translation."

    if not source_language or not target_language:
        yield "Both source and target language codes are required."

    try:
        translate = boto3.client(service_name='translate', region_name=region_name, use_ssl=True)
        result = translate.translate_text(Text=data, SourceLanguageCode=source_language, TargetLanguageCode=target_language)
        yield result.get('TranslatedText', 'No translation found.')

    except BotoCoreError as e:
        logging.info(f"BotoCoreError occurred: {e}")
        yield "Error with AWS Translate service configuration or request."

    except ClientError as e:
        logging.info(f"ClientError occurred: {e}")
        yield "Error with AWS client or network issue."


import hashlib
from .create_user import create_user

async def create_default_user():
    default_user_email = "default_user@example.com"
    default_user_password = "default_password"

    user = await create_user(
        email = default_user_email,
        password = await hash_password(default_user_password),
        is_superuser = True,
        is_active = True,
        is_verified = True,
        auto_login = True,
    )

    return user

async def hash_password(password: str) -> str:
    return hashlib.sha256(password.encode()).hexdigest()



""" This module is used to set the configuration of the system."""
import os
from cognee.base_config import get_base_config
from cognee.modules.cognify.config import get_cognify_config
from cognee.infrastructure.data.chunking.config import get_chunk_config
from cognee.infrastructure.databases.vector import get_vectordb_config
from cognee.infrastructure.databases.graph.config import get_graph_config
from cognee.infrastructure.databases.relational import get_relational_config
from cognee.infrastructure.files.storage import LocalStorage

class config():
    @staticmethod
    def system_root_directory(system_root_directory: str):
        databases_directory_path = os.path.join(system_root_directory, "databases")

        relational_config = get_relational_config()
        relational_config.db_path = databases_directory_path
        LocalStorage.ensure_directory_exists(databases_directory_path)

        graph_config = get_graph_config()
        graph_config.graph_file_path = os.path.join(databases_directory_path, "cognee.graph")

        vector_config = get_vectordb_config()
        if vector_config.vector_engine_provider == "lancedb":
            vector_config.vector_db_url = os.path.join(databases_directory_path, "cognee.lancedb")

    @staticmethod
    def data_root_directory(data_root_directory: str):
        base_config = get_base_config()
        base_config.data_root_directory = data_root_directory

    @staticmethod
    def monitoring_tool(monitoring_tool: object):
        base_config = get_base_config()
        base_config.monitoring_tool = monitoring_tool

    @staticmethod
    def set_classification_model(classification_model: object):
        cognify_config = get_cognify_config()
        cognify_config.classification_model =  classification_model

    @staticmethod
    def set_summarization_model(summarization_model: object):
        cognify_config = get_cognify_config()
        cognify_config.summarization_model=summarization_model

    @staticmethod
    def set_graph_model(graph_model: object):
        graph_config = get_graph_config()
        graph_config.graph_model = graph_model

    @staticmethod
    def set_graph_database_provider(graph_database_provider: str):
        graph_config = get_graph_config()
        graph_config.graph_database_provider = graph_database_provider

    @staticmethod
    def llm_provider(llm_provider: str):
        graph_config = get_graph_config()
        graph_config.llm_provider = llm_provider

    @staticmethod
    def llm_endpoint(llm_endpoint: str):
        graph_config = get_graph_config()
        graph_config.llm_endpoint = llm_endpoint

    @staticmethod
    def llm_model(llm_model: str):
        graph_config = get_graph_config()
        graph_config.llm_model = llm_model

    @staticmethod
    def set_chunk_strategy(chunk_strategy: object):
        chunk_config = get_chunk_config()
        chunk_config.chunk_strategy = chunk_strategy

    @staticmethod
    def set_chunk_engine(chunk_engine: object):
        chunk_config = get_chunk_config()
        chunk_config.chunk_engine = chunk_engine

    @staticmethod
    def set_chunk_overlap(chunk_overlap: object):
        chunk_config = get_chunk_config()
        chunk_config.chunk_overlap = chunk_overlap

    @staticmethod
    def set_chunk_size(chunk_size: object):
        chunk_config = get_chunk_config()
        chunk_config.chunk_size = chunk_size


    @staticmethod
    def set_vector_engine_provider(vector_engine_provider: str):
        vector_db_config = get_vectordb_config()
        vector_db_config.vector_engine_provider = vector_engine_provider

    @staticmethod
    def set_vector_db_key(db_key: str):
        vector_db_config = get_vectordb_config()
        vector_db_config.vector_db_key = db_key


    @staticmethod
    def set_vector_db_url(db_url: str):
        vector_db_config = get_vectordb_config()
        vector_db_config.vector_db_url = db_url

    @staticmethod
    def set_graphistry_username(graphistry_username: str):
        base_config = get_base_config()
        base_config.graphistry_username = graphistry_username

    @staticmethod
    def set_graphistry_password(graphistry_password: str):
        base_config = get_base_config()
        base_config.graphistry_password = graphistry_password






from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.llm.prompts import  render_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client

async def categorize_relevant_summary(query: str, summaries, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    enriched_query= render_prompt("categorize_summary.txt", {"query": query, "summaries": summaries})

    system_prompt = "Choose the relevant summaries and return appropriate output based on the model"

    llm_output = await llm_client.acreate_structured_output(enriched_query, system_prompt, response_model)

    return llm_output



from .ModelBase import Base
from .DatabaseEngine import DatabaseEngine
from .sqlite.SqliteEngine import SqliteEngine
from .duckdb.DuckDBAdapter import DuckDBAdapter
from .config import get_relational_config
from .create_db_and_tables import create_db_and_tables
from .get_relational_engine import get_relational_engine



from cognee.infrastructure.databases.vector import get_vector_engine

async def search_similarity(query: str) -> list[str, str]:
    """
    Parameters:
    - query (str): The query string to filter nodes by.

    Returns:
    - list(chunk): A list of objects providing information about the chunks related to query.
    """
    vector_engine = get_vector_engine()

    similar_results = await vector_engine.search("chunks", query, limit = 5)

    results = [
        parse_payload(result.payload) for result in similar_results
    ]

    return results


def parse_payload(payload: dict) -> dict:
    return {
        "text": payload["text"],
        "chunk_id": payload["chunk_id"],
        "document_id": payload["document_id"],
    }



from typing import Type, Optional, get_args, get_origin
from pydantic import BaseModel
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

async def add_model_class_to_graph(
    model_class: Type[BaseModel],
    graph: GraphDBInterface,
    parent: Optional[str] = None,
    relationship: Optional[str] = None,
):
    model_name = model_class.__name__

    if await graph.extract_node(model_name):
        return

    await graph.add_node(model_name, dict(type = "model"))

    if parent and relationship:
        await graph.add_edge(
            parent,
            model_name,
            relationship,
            dict(
                relationship_name = relationship,
                source_node_id = parent,
                target_node_id = model_name,
            ),
        )

    for field_name, field in model_class.model_fields.items():
        original_types = get_args(field.annotation)
        field_type = original_types[0] if len(original_types) > 0 else None

        if field_type is None:
            continue

        if hasattr(field_type, "model_fields"):  # Check if field type is a Pydantic model
            await add_model_class_to_graph(field_type, graph, model_name, field_name)
        elif get_origin(field.annotation) == list:
            list_types = get_args(field_type)
            for item_type in list_types:
                await add_model_class_to_graph(item_type, graph, model_name, field_name)
        elif isinstance(field_type, list):
            item_type = get_args(field_type)[0]
            if hasattr(item_type, "model_fields"):
                await add_model_class_to_graph(item_type, graph, model_name, field_name)
            else:
                await graph.add_node(str(item_type), dict(type = "value"))
                await graph.add_edge(
                    model_name,
                    str(item_type),
                    field_name,
                    dict(
                        relationship_name = field_name,
                        source_node_id = model_name,
                        target_node_id = str(item_type),
                    ),
                )
        else:
            await graph.add_node(str(field_type), dict(type = "value"))
            await graph.add_edge(
                model_name,
                str(field_type),
                field_name,
                dict(
                    relationship_name = field_name,
                    source_node_id = model_name,
                    target_node_id = str(field_type),
                ),
            )
